import re
from html.parser import HTMLParser
import json
import warnings

from django.core.exceptions import ValidationError
from django.contrib.staticfiles.finders import find
from django.forms import Media, widgets
from django.utils.html import escape, format_html, format_html_join
from django.utils.translation import gettext_lazy as _


class JSONMultiWidget(widgets.MultiWidget):
    """
    Deprecated.
    Base class for MultiWidgets using a JSON field in database.
    """
    html_parser = HTMLParser()

    def __init__(self, glossary_fields):
        from cmsplugin_cascade.fields import GlossaryField
        warnings.warn("JSONMultiWidget is deprecated")

        self.glossary_fields = list(glossary_fields)
        self.normalized_fields = []
        for field in self.glossary_fields:
            if isinstance(field, GlossaryField):
                self.normalized_fields.append(field)
            elif isinstance(field, (list, tuple)):
                self.normalized_fields.extend([f for f in field if isinstance(f, GlossaryField)])
            else:
                raise ValueError("Given fields must be of type GlossaryField or list of thereof")
        unique_keys = set([field.name for field in self.normalized_fields])
        if len(self.normalized_fields) > len(unique_keys):
            raise ValueError('List of glossary_fields may contain only unique keys')
        super().__init__((field.widget for field in self.normalized_fields))

    def decompress(self, values):
        if not isinstance(values, dict):
            values = json.loads(values or '{}')
        for field in self.normalized_fields:
            initial_value = field.initial() if callable(field.initial) else field.initial
            if isinstance(field.widget, widgets.MultiWidget):
                values[field.name] = field.widget.decompress(values.get(field.name, initial_value))
            else:
                values.setdefault(field.name, initial_value)
        return values

    def value_from_datadict(self, data, files, name):
        result = {}
        for field in self.normalized_fields:
            if isinstance(field.widget, widgets.MultiWidget):
                result[field.name] = field.widget.value_from_datadict(data, files, field.name)
            elif getattr(field.widget, 'allow_multiple_selected', False):
                result[field.name] = list(map(escape, data.getlist(field.name)))
            else:
                result[field.name] = escape(data.get(field.name, ''))
        return result

    def value_omitted_from_data(self, data, files, name):
        # required since Django-1.10 during invocation of `construct_instance`
        return self.normalized_fields and all(
            field.widget.value_omitted_from_data(data, files, field.name)
            for field in self.normalized_fields
        )

    def render(self, name, value, attrs=None, renderer=None):
        values = self.decompress(value)
        render_fieldsets = []
        for fieldset in self.glossary_fields:
            render_fields = []
            if not isinstance(fieldset, (list, tuple)):
                fieldset = [fieldset]
            for field in fieldset:
                field_attrs = dict(**attrs)
                field_attrs.update(id='{id}_{0}'.format(field.name, **attrs))
                field_value = values.get(field.name)
                if isinstance(field_value, str):
                    field_value = self.html_parser.unescape(field_value)
                render_fields.append((
                    field.name,
                    str(field.label),
                    field.widget.render(field.name, field_value, field_attrs),
                    str(field.help_text),
                ))
            html = format_html_join('',
                 '<div class="glossary-field glossary_{0}"><h1>{1}</h1><div class="glossary-box">{2}</div><small>{3}</small></div>',
                 render_fields)
            render_fieldsets.append((html,))
        return format_html_join('\n', '<div class="glossary-widget">{0}</div>', render_fieldsets)


class NumberInputWidget(widgets.NumberInput):
    validation_pattern = re.compile('^-?\d+(\.\d{1,2})?$')
    required_message = _("In '%(label)s': This field is required.")
    invalid_message = _("In '%(label)s': Value '%(value)s' shall contain a valid decimal number.")

    def validate(self, value):
        if value and not self.validation_pattern.match(value):
            raise ValidationError(self.invalid_message, code='invalid', params={'value': value})


class AColorPickerMixin:
    acolorpicker_js = 'node_modules/a-color-picker/dist/acolorpicker.js'

    def __init__(self, with_alpha, *args, **kwargs):
        self.with_acolorpicker = bool(find(self.acolorpicker_js))
        if with_alpha and not self.with_acolorpicker:
            msg = "Node package 'a-color-picker' not found in 'node_modules'.\n" \
                  "Please install it from npm into your project directory and add " \
                  "('node_modules', 'project_directory/node_modules') to your STATICFILES_DIRS."
            raise FileNotFoundError(msg)
        super().__init__(*args, **kwargs)

    @property
    def media(self):
        if self.with_acolorpicker:
            js = [self.acolorpicker_js, 'admin/js/jquery.init.js', 'cascade/js/admin/colorpicker.js']
        else:
            js = ['admin/js/jquery.init.js', 'cascade/js/admin/colorpicker.js']
        return Media(js=js)

    @classmethod
    def rgb2hex(cls, val):
        match = re.search('rgb[a]?\((\d+),\s*(\d+),\s*(\d+)[^)]*\)', val)
        if match:
            val = "#{:02x}{:02x}{:02x}".format(*[int(m) for m in match.groups()])
        return val


class ColorPickerWidget(AColorPickerMixin, widgets.MultiWidget):
    """
    Use this field to enter a color value. Clicking onto this widget will pop up a color picker.
    The value passed to the consumer is a tuple of a Boolean and a string guaranteed to be in #rgb format.
    """
    template_name = 'cascade/admin/widgets/colorpicker.html'

    def __init__(self, with_alpha):
        widget_list = [
            widgets.TextInput(attrs={'data-with_alpha': str(with_alpha).lower(), 'class': 'cascade-rgba'}),
            widgets.TextInput(attrs={'type': 'color'}),
            widgets.CheckboxInput(),
        ]
        super().__init__(with_alpha, widget_list)

    class Media:
        css = {'all': ['cascade/css/admin/colorpicker.css']}

    def decompress(self, values):
        assert isinstance(values, (list, tuple)), "Values to decompress are kept as lists in JSON"
        values = list(values)
        return values

    def value_from_datadict(self, data, files, name):
        if self.with_acolorpicker:
            color = data.get('{}_0'.format(name))
        else:
            color = data.get('{}_1'.format(name))
        values = [
            escape(color),
            bool(data.get('{}_2'.format(name))),
        ]
        return values

    def get_context(self, name, value, attrs):
        if not isinstance(value, list):
            values = self.decompress(value)
        else:
            values = list(value)
        assert len(values) == 2
        values.insert(0, values[0])
        values[1] = self.rgb2hex(values[0])
        context = super().get_context(name, values, attrs)
        return context


class MultipleTextInputWidget(widgets.MultiWidget):
    """
    A widgets accepting multiple input values.
    """
    required = False

    def __init__(self, labels, attrs=None):
        text_widgets = [widgets.TextInput()] * len(labels)
        super().__init__(text_widgets, attrs)
        self.labels = labels[:]

    def decompress(self, values):
        assert isinstance(values, dict), "Values to decompress are kept as dict in JSON"
        return list(values.values())

    def value_from_datadict(self, data, files, name):
        values = [escape(data.get('{0}-{1}'.format(name, label), '')) for label in self.labels]
        return values

    def render(self, name, value, attrs=None, renderer=None):
        widgets = []
        values = value[:] if isinstance(value, (list, tuple)) else []
        values.extend([''] * max(len(self.labels) - len(values), 0))
        elem_id = attrs['id']
        for index, key in enumerate(self.labels):
            label = '{0}-{1}'.format(name, key)
            attrs['id'] = '{0}_{1}'.format(elem_id, key)
            widgets.append((self.widgets[index].render(label, values[index], attrs, renderer), key, label))
        return format_html('<div>{0}</div>',
                           format_html_join('\n',
                                            '<div class="sibling-field"><label for="{2}">{1}</label>{0}</div>',
                                            widgets))


class BorderChoiceWidget(AColorPickerMixin, widgets.MultiWidget):
    """
    Use this field to enter the three values of a border: width style color.
    """
    template_name = 'cascade/admin/widgets/borderchoice.html'

    def __init__(self, choices, with_alpha):
        widget_list = [
            widgets.TextInput(attrs={'size': 5}),
            widgets.Select(choices=choices),
            widgets.TextInput(attrs={'data-with_alpha': str(with_alpha).lower(), 'class': 'cascade-rgba'}),
            widgets.TextInput(attrs={'type': 'color'}),
        ]
        super().__init__(with_alpha, widget_list)

    @property
    def media(self):
        return super().media + Media(css={'all': ['cascade/css/admin/colorpicker.css',
                                                  'cascade/css/admin/borderchoice.css']})

    def decompress(self, values):
        assert isinstance(values, (list, tuple)), "Values to decompress are kept as lists in JSON"
        return list(values)

    def value_from_datadict(self, data, files, name):
        try:
            if self.with_acolorpicker:
                color = data.get('{}_2'.format(name))
            else:
                color = data.get('{}_3'.format(name))
            values = [
                escape(data['{}_0'.format(name)]),
                escape(data['{}_1'.format(name)]),
                escape(color),
            ]
        except KeyError:
            values = ['0px', 'none', 'none']
        return values

    def get_context(self, name, value, attrs):
        if not isinstance(value, list):
            values = self.decompress(value)
        else:
            values = list(value)
        assert len(values) == 3
        values.insert(2, values[2])
        values[3] = self.rgb2hex(values[2])
        context = super().get_context(name, values, attrs)
        return context



import logging
from django.forms import widgets, ChoiceField, MultipleChoiceField
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.bootstrap4.grid import Breakpoint
from cmsplugin_cascade.bootstrap4.utils import get_image_tags, IMAGE_RESIZE_OPTIONS, IMAGE_SHAPE_CHOICES
from cmsplugin_cascade.image import ImageFormMixin, ImagePropertyMixin
from cmsplugin_cascade.fields import SizeField
from cmsplugin_cascade.link.config import LinkPluginBase, LinkFormMixin
from cmsplugin_cascade.link.plugin_base import LinkElementMixin

logger = logging.getLogger('cascade.bootstrap4')


class BootstrapImageFormMixin(ImageFormMixin):
    ALIGNMENT_OPTIONS = [
        ('float-left', _("Left")),
        ('float-right', _("Right")),
        ('mx-auto', _("Center")),
    ]

    image_shapes = MultipleChoiceField(
        label=_("Image Shapes"),
        choices=IMAGE_SHAPE_CHOICES,
        widget=widgets.CheckboxSelectMultiple,
        initial=['img-fluid']
    )

    image_width_responsive = SizeField(
        label=_("Responsive Image Width"),
        allowed_units=['%'],
        initial='100%',
        required = False,
        help_text=_("Set the image width in percent relative to containing element."),
    )

    image_width_fixed = SizeField(
        label=_("Fixed Image Width"),
        allowed_units=['px'],
        required = False,
        help_text=_("Set a fixed image width in pixels."),
    )

    image_height = SizeField(
        label=_("Adapt Image Height"),
        allowed_units=['px', '%'],
        required = False,
        help_text=_("Set a fixed height in pixels, or percent relative to the image width."),
    )

    resize_options = MultipleChoiceField(
        label=_("Resize Options"),
        choices=IMAGE_RESIZE_OPTIONS,
        widget=widgets.CheckboxSelectMultiple,
        required = False,
        help_text=_("Options to use when resizing the image."),
        initial=['subject_location', 'high_resolution'],
    )

    image_alignment = ChoiceField(
        label=_("Image Alignment"),
        choices=ALIGNMENT_OPTIONS,
        widget=widgets.RadioSelect,
        required = False,
        help_text=_("How to align a non-responsive image."),
    )

    class Meta:
        entangled_fields = {'glossary': ['image_shapes', 'image_width_responsive', 'image_width_fixed',
                                         'image_height', 'resize_options', 'image_alignment']}


class BootstrapImagePlugin(LinkPluginBase):
    name = _("Image")
    module = 'Bootstrap'
    parent_classes = ['BootstrapColumnPlugin']
    require_parent = True
    allow_children = False
    raw_id_fields = LinkPluginBase.raw_id_fields + ['image_file']
    model_mixins = (ImagePropertyMixin, LinkElementMixin,)
    admin_preview = False
    ring_plugin = 'ImagePlugin'
    form = type('BootstrapImageForm', (LinkFormMixin, BootstrapImageFormMixin), {'require_link': False})
    render_template = 'cascade/bootstrap4/linked-image.html'
    default_css_attributes = ['image_shapes', 'image_alignment']
    html_tag_attributes = {'image_title': 'title', 'alt_tag': 'tag'}
    html_tag_attributes.update(LinkPluginBase.html_tag_attributes)

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/imageplugin.js']

    def render(self, context, instance, placeholder):
        context = self.super(BootstrapImagePlugin, self).render(context, instance, placeholder)
        try:
            image_tags = get_image_tags(instance)
        except Exception as exc:
            logger.warning("Unable generate image tags. Reason: {}".format(exc))
        else:
            extra_styles = image_tags.pop('extra_styles', None)
            if extra_styles:
                inline_styles = instance.glossary.get('inline_styles', {})
                inline_styles.update(extra_styles)
                instance.glossary['inline_styles'] = inline_styles
            context.update(dict(**image_tags))
        return context

    @classmethod
    def get_css_classes(cls, obj):
        css_classes = cls.super(BootstrapImagePlugin, cls).get_css_classes(obj)
        css_class = obj.glossary.get('css_class')
        if css_class:
            css_classes.append(css_class)
        return css_classes

    @classmethod
    def get_identifier(cls, obj):
        try:
            content = str(obj.image)
        except AttributeError:
            content = _("No Image")
        return mark_safe(content)

    @classmethod
    def sanitize_model(cls, obj):
        sanitized = False
        parent = obj.parent
        try:
            while parent.plugin_type != 'BootstrapColumnPlugin':
                parent = parent.parent
            grid_column = parent.get_bound_plugin().get_grid_instance()
            min_max_bounds = grid_column.get_min_max_bounds()
            if obj.glossary.get('column_bounds') != min_max_bounds:
                obj.glossary['column_bounds'] = min_max_bounds
                sanitized = True
            obj.glossary.setdefault('media_queries', {})
            for bp in Breakpoint:
                media_query = '{} {:.2f}px'.format(bp.media_query, grid_column.get_bound(bp).max)
                if obj.glossary['media_queries'].get(bp.name) != media_query:
                    obj.glossary['media_queries'][bp.name] = media_query
                    sanitized = True
        except AttributeError:
            logger.warning("ImagePlugin(pk={}) has no ColumnPlugin as ancestor.".format(obj.pk))
            return
        return sanitized

plugin_pool.register_plugin(BootstrapImagePlugin)



import re
import logging
from django.forms import widgets
from django.forms.fields import IntegerField, MultipleChoiceField
from django.utils.safestring import mark_safe
from django.utils.translation import ngettext_lazy, gettext_lazy as _

from entangled.forms import EntangledModelFormMixin
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.bootstrap4.fields import BootstrapMultiSizeField
from cmsplugin_cascade.bootstrap4.grid import Breakpoint
from cmsplugin_cascade.bootstrap4.picture import get_picture_elements
from cmsplugin_cascade.bootstrap4.plugin_base import BootstrapPluginBase
from cmsplugin_cascade.bootstrap4.utils import IMAGE_RESIZE_OPTIONS
from cmsplugin_cascade.forms import ManageChildrenFormMixin
from cmsplugin_cascade.image import ImagePropertyMixin, ImageFormMixin

logger = logging.getLogger('cascade')


class CarouselSlidesFormMixin(ManageChildrenFormMixin, EntangledModelFormMixin):
    OPTION_CHOICES = [('slide', _("Animate")), ('pause', _("Pause")), ('wrap', _("Wrap"))]

    num_children = IntegerField(min_value=1, initial=1,
        label=_('Slides'),
        help_text=_('Number of slides for this carousel.'),
    )

    interval = IntegerField(
        label=_("Interval"),
        initial=5,
        help_text=_("Change slide after this number of seconds."),
    )

    options = MultipleChoiceField(
        label=_('Options'),
        choices=OPTION_CHOICES,
        widget=widgets.CheckboxSelectMultiple,
        initial=['slide', 'wrap', 'pause'],
        help_text=_("Adjust interval for the carousel."),
    )

    container_max_heights = BootstrapMultiSizeField(
        label=_("Carousel heights"),
        allowed_units=['px'],
        initial=['100px', '150px', '200px', '250px', '300px'],
        help_text=_("Heights of Carousel in pixels for distinct Bootstrap's breakpoints."),
    )

    resize_options = MultipleChoiceField(
        label=_("Resize Options"),
        choices=IMAGE_RESIZE_OPTIONS,
        widget=widgets.CheckboxSelectMultiple,
        help_text=_("Options to use when resizing the image."),
        initial=['upscale', 'crop', 'subject_location', 'high_resolution'],
    )

    class Meta:
        untangled_fields = ['num_children']
        entangled_fields = {'glossary': ['interval', 'options', 'container_max_heights', 'resize_options']}


class BootstrapCarouselPlugin(BootstrapPluginBase):
    name = _("Carousel")
    default_css_class = 'carousel'
    default_css_attributes = ['options']
    parent_classes = ['BootstrapColumnPlugin']
    render_template = 'cascade/bootstrap4/{}carousel.html'
    default_inline_styles = {'overflow': 'hidden'}
    form = CarouselSlidesFormMixin
    DEFAULT_CAROUSEL_ATTRIBUTES = {'data-ride': 'carousel'}

    @classmethod
    def get_identifier(cls, obj):
        num_cols = obj.get_num_children()
        content = ngettext_lazy('with {0} slide', 'with {0} slides', num_cols).format(num_cols)
        return mark_safe(content)

    @classmethod
    def get_css_classes(cls, obj):
        css_classes = cls.super(BootstrapCarouselPlugin, cls).get_css_classes(obj)
        if 'slide' in obj.glossary.get('options', []):
            css_classes.append('slide')
        return css_classes

    @classmethod
    def get_html_tag_attributes(cls, obj):
        attributes = cls.super(BootstrapCarouselPlugin, cls).get_html_tag_attributes(obj)
        attributes.update(cls.DEFAULT_CAROUSEL_ATTRIBUTES)
        attributes['data-interval'] = 1000 * int(obj.glossary.get('interval', 5))
        options = obj.glossary.get('options', [])
        attributes['data-pause'] = 'pause' in options and 'hover' or 'false'
        attributes['data-wrap'] = 'wrap' in options and 'true' or 'false'
        return attributes

    def save_model(self, request, obj, form, change):
        wanted_children = int(form.cleaned_data.get('num_children'))
        super().save_model(request, obj, form, change)
        self.extend_children(obj, wanted_children, BootstrapCarouselSlidePlugin)
        obj.sanitize_children()

    @classmethod
    def sanitize_model(cls, obj):
        sanitized = super().sanitize_model(obj)
        complete_glossary = obj.get_complete_glossary()
        # fill all invalid heights for this container to a meaningful value
        max_height = max(obj.glossary['container_max_heights'].values())
        pattern = re.compile(r'^(\d+)px$')
        for bp in complete_glossary.get('breakpoints', ()):
            if not pattern.match(obj.glossary['container_max_heights'].get(bp, '')):
                obj.glossary['container_max_heights'][bp] = max_height
        return sanitized

plugin_pool.register_plugin(BootstrapCarouselPlugin)


class BootstrapCarouselSlidePlugin(BootstrapPluginBase):
    name = _("Slide")
    model_mixins = (ImagePropertyMixin,)
    default_css_class = 'img-fluid'
    parent_classes = ['BootstrapCarouselPlugin']
    raw_id_fields = ['image_file']
    html_tag_attributes = {'image_title': 'title', 'alt_tag': 'tag'}
    render_template = 'cascade/bootstrap4/carousel-slide.html'
    form = ImageFormMixin
    alien_child_classes = True

    def render(self, context, instance, placeholder):
        context = self.super(BootstrapCarouselSlidePlugin, self).render(context, instance, placeholder)
        # slide image shall be rendered in a responsive context using the ``<picture>`` element
        try:
            parent_glossary = instance.parent.get_bound_plugin().glossary
            instance.glossary.update(responsive_heights=parent_glossary['container_max_heights'])
            elements = get_picture_elements(instance)
        except Exception as exc:
            logger.warning("Unable generate picture elements. Reason: {}".format(exc))
        else:
            context.update({
                'is_fluid': False,
                'elements': elements,
            })
        return context

    @classmethod
    def sanitize_model(cls, obj):
        sanitized = super().sanitize_model(obj)
        resize_options = obj.get_parent_glossary().get('resize_options', [])
        if obj.glossary.get('resize_options') != resize_options:
            obj.glossary.update(resize_options=resize_options)
            sanitized = True
        parent = obj.parent
        while parent.plugin_type != 'BootstrapColumnPlugin':
            parent = parent.parent
            if parent is None:
                logger.warning("PicturePlugin(pk={}) has no ColumnPlugin as ancestor.".format(obj.pk))
                return
        grid_column = parent.get_bound_plugin().get_grid_instance()
        obj.glossary.setdefault('media_queries', {})
        for bp in Breakpoint:
            obj.glossary['media_queries'].setdefault(bp.name, {})
            width = round(grid_column.get_bound(bp).max)
            if obj.glossary['media_queries'][bp.name].get('width') != width:
                obj.glossary['media_queries'][bp.name]['width'] = width
                sanitized = True
            if obj.glossary['media_queries'][bp.name].get('media') != bp.media_query:
                obj.glossary['media_queries'][bp.name]['media'] = bp.media_query
                sanitized = True
        return sanitized

    @classmethod
    def get_identifier(cls, obj):
        try:
            content = obj.image.name or obj.image.original_filename
        except AttributeError:
            content = _("Empty Slide")
        return mark_safe(content)

plugin_pool.register_plugin(BootstrapCarouselSlidePlugin)



from urllib.parse import urlparse
import requests

from django.contrib import admin
from django.contrib.sites.shortcuts import get_current_site
from django.forms import Media, widgets
from django.db.models import Q
from django.http import JsonResponse, HttpResponseForbidden, HttpResponseNotFound
from django.urls import re_path, reverse
from django.utils.translation import get_language_from_request

from cms.models.pagemodel import Page
from cms.extensions import PageExtensionAdmin
from cms.utils.page import get_page_from_path
from cmsplugin_cascade.models import CascadePage, IconFont
from cmsplugin_cascade.link.forms import format_page_link


@admin.register(CascadePage)
class CascadePageAdmin(PageExtensionAdmin):
    add_form_template = change_form_template = 'cascade/admin/change_form.html'
    fields = ['icon_font', 'menu_symbol']

    @property
    def media(self):
        media = super().media
        media += Media(css={'all': ['cascade/css/admin/cascadepage.css']},
                       js=['admin/js/jquery.init.js', 'cascade/js/admin/cascadepage.js'])
        return media

    def get_form(self, request, obj=None, **kwargs):
        options = dict(kwargs, widgets={'menu_symbol': widgets.HiddenInput})
        ModelForm = super().get_form(request, obj, **options)
        return ModelForm

    def get_urls(self):
        urls = [
            re_path(r'^get_page_sections/$', lambda _: JsonResponse({'element_ids': []}),
                name='get_page_sections'),  # just to reverse
            re_path(r'^get_page_sections/(?P<page_pk>\d+)$',
                self.admin_site.admin_view(self.get_page_sections)),
            re_path(r'^published_pages/$', self.get_published_pagelist, name='get_published_pagelist'),
            re_path(r'^fetch_fonticons/(?P<iconfont_id>[0-9]+)$', self.fetch_fonticons),
            re_path(r'^fetch_fonticons/$', self.fetch_fonticons, name='fetch_fonticons'),
            re_path(r'^validate_exturl/$', self.validate_exturl, name='validate_exturl'),
        ]
        urls.extend(super().get_urls())
        return urls

    def get_page_sections(self, request, page_pk=None):
        choices = []
        try:
            extended_glossary = self.model.objects.get(extended_object_id=page_pk).glossary
            for key, val in extended_glossary['element_ids'].items():
                choices.append((key, val))
        except (self.model.DoesNotExist, KeyError):
            pass
        return JsonResponse({'element_ids': choices})

    def get_published_pagelist(self, request, *args, **kwargs):
        """
        This view is used by the SearchLinkField as the user types to feed the autocomplete drop-down.
        """
        if not request.is_ajax():
            return HttpResponseForbidden()
        data = {'results': []}
        language = get_language_from_request(request)
        query_term = request.GET.get('term')
        if not query_term:
            return JsonResponse(data)

        # first, try to resolve by URL if it points to a local resource
        parse_result = urlparse(query_term)
        if parse_result.netloc.split(':')[0] == request.META['HTTP_HOST'].split(':')[0]:
            site = get_current_site(request)
            path = parse_result.path.lstrip(reverse('pages-root')).rstrip('/')
            page = get_page_from_path(site, path)
            if page:
                data['results'].append(self.get_result_set(language, page))
                return JsonResponse(data)

        # otherwise resolve by search term
        matching_published_pages = Page.objects.published().public().filter(
            Q(title_set__title__icontains=query_term, title_set__language=language)
            | Q(title_set__path__icontains=query_term, title_set__language=language)
            | Q(title_set__menu_title__icontains=query_term, title_set__language=language)
            | Q(title_set__page_title__icontains=query_term, title_set__language=language)
        ).distinct().order_by('title_set__title').iterator()

        for page in matching_published_pages:
            data['results'].append(self.get_result_set(language, page))
            if len(data['results']) > 15:
                break
        return JsonResponse(data)

    def get_result_set(self, language, page):
        title = page.get_title(language=language)
        path = page.get_absolute_url(language=language)
        return {
            'id': page.pk,
            'text': format_page_link(title, path),
        }

    def fetch_fonticons(self, request, iconfont_id=None):
        try:
            icon_font = IconFont.objects.get(id=iconfont_id)
        except IconFont.DoesNotExist:
            return HttpResponseNotFound("IconFont with id={} does not exist".format(iconfont_id))
        else:
            data = dict(icon_font.config_data)
            data.pop('glyphs', None)
            data['families'] = icon_font.get_icon_families()
            return JsonResponse(data)

    def validate_exturl(self, request):
        """
        Perform a GET request onto the given external URL and return its status.
        """
        exturl = request.GET.get('exturl')
        request_headers = {'User-Agent': 'Django-CMS-Cascade'}
        try:
            response = requests.get(exturl, allow_redirects=True, headers=request_headers)
        except Exception:
            return JsonResponse({'status_code': 500})
        else:
            return JsonResponse({'status_code': response.status_code})

    def changeform_view(self, request, object_id=None, form_url='', extra_context=None):
        extra_context = dict(extra_context or {}, icon_fonts=IconFont.objects.all())
        return super().changeform_view(
             request, object_id=object_id, form_url=form_url, extra_context=extra_context)



from django.forms import widgets, ChoiceField, MultipleChoiceField
from django.utils.html import format_html_join
from django.utils.translation import gettext_lazy as _

from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.fields import SizeField
from cmsplugin_cascade.image import ImageFormMixin, ImagePropertyMixin
from cmsplugin_cascade.link.config import LinkPluginBase, LinkFormMixin
from cmsplugin_cascade.link.plugin_base import LinkElementMixin
from cmsplugin_cascade.utils import compute_aspect_ratio


class TextImageFormMixin(ImageFormMixin):
    RESIZE_OPTIONS = [
        ('upscale', _("Upscale image")),
        ('crop', _("Crop image")),
        ('subject_location', _("With subject location")),
        ('high_resolution', _("Optimized for Retina")),
    ]

    image_width = SizeField(
        label=_("Image Width"),
        allowed_units=['px'],
        required=True,
        help_text=_("Set the image width in pixels."),
    )

    image_height = SizeField(
        label=_("Image Height"),
        allowed_units=['px'],
        required=False,
        help_text=_("Set the image height in pixels."),
    )

    resize_options = MultipleChoiceField(
        label=_("Resize Options"),
        choices = RESIZE_OPTIONS,
        required=False,
        widget=widgets.CheckboxSelectMultiple,
        help_text=_("Options to use when resizing the image."),
        initial=['subject_location', 'high_resolution']
    )

    alignement = ChoiceField(
        label=_("Alignement"),
        choices=[('', _("Not aligned")), ('left', _("Left")), ('right', _("Right"))],
        required=False,
        widget=widgets.RadioSelect,
        initial='',
    )

    class Meta:
        entangled_fields = {'glossary': ['image_width', 'image_height', 'resize_options', 'alignement']}


class TextImagePlugin(LinkPluginBase):
    name = _("Image in text")
    text_enabled = True
    ring_plugin = 'TextImagePlugin'
    render_template = 'cascade/plugins/textimage.html'
    parent_classes = ['TextPlugin']
    model_mixins = (ImagePropertyMixin, LinkElementMixin)
    allow_children = False
    require_parent = False
    form = type('TextImageForm', (LinkFormMixin, TextImageFormMixin), {'require_link': False})
    html_tag_attributes = {'image_title': 'title', 'alt_tag': 'tag'}
    html_tag_attributes.update(LinkPluginBase.html_tag_attributes)

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/textimageplugin.js']

    @classmethod
    def requires_parent_plugin(cls, slot, page):
        """
        Workaround for `PluginPool.get_all_plugins()`, otherwise TextImagePlugin is not allowed
        as a child of a `TextPlugin`.
        """
        return False

    @classmethod
    def get_inline_styles(cls, instance):
        inline_styles = cls.super(TextImagePlugin, cls).get_inline_styles(instance)
        alignement = instance.glossary.get('alignement')
        if alignement:
            inline_styles['float'] = alignement
        return inline_styles

    def render(self, context, instance, placeholder):
        context = self.super(TextImagePlugin, self).render(context, instance, placeholder)
        try:
            aspect_ratio = compute_aspect_ratio(instance.image)
        except Exception:
            # if accessing the image file fails, abort here
            return context
        resize_options = instance.glossary.get('resize_options', {})
        crop = 'crop' in resize_options
        upscale = 'upscale' in resize_options
        subject_location = instance.image.subject_location if 'subject_location' in resize_options else False
        high_resolution = 'high_resolution' in resize_options
        image_width = instance.glossary.get('image_width', '')
        if not image_width.endswith('px'):
            return context
        image_width = int(image_width.rstrip('px'))
        image_height = instance.glossary.get('image_height', '')
        if image_height.endswith('px'):
            image_height = int(image_height.rstrip('px'))
        else:
            image_height = int(round(image_width * aspect_ratio))
        context['src'] = {
            'size': (image_width, image_height),
            'size2x': (image_width * 2, image_height * 2),
            'crop': crop,
            'upscale': upscale,
            'subject_location': subject_location,
            'high_resolution': high_resolution,
        }
        link_attributes = LinkPluginBase.get_html_tag_attributes(instance)
        context['link_html_tag_attributes'] = format_html_join(' ', '{0}="{1}"',
            [(attr, val) for attr, val in link_attributes.items() if val]
        )
        return context

plugin_pool.register_plugin(TextImagePlugin)



from django.contrib import admin
from django.utils.translation import gettext as _
from django.utils.html import format_html
from entangled.forms import EntangledModelForm
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.models import IconFont
from cmsplugin_cascade.plugin_base import CascadePluginMixinMetaclass
from cmsplugin_cascade.models import SharedGlossary, SharableCascadeElement


@admin.register(SharedGlossary)
class SharedGlossaryAdmin(admin.ModelAdmin):
    change_form_template = 'cascade/admin/sharedglossary_change_form.html'
    list_display = ['identifier', 'plugin_name', 'used_by']
    readonly_fields = ['plugin_name']
    list_filter = ['plugin_type']

    class Media:
        css = {'all': ['cascade/css/admin/partialfields.css', 'cascade/css/admin/editplugin.css']}

    def get_form(self, request, obj=None, **kwargs):
        """
        Creates a temporary form with an identifier and the fields declared as sharables for the
        corresponding plugin model. Additionally it adds dynamic fields to edit the content inside
        the model field `glossary`. The layout, validation and media files for these dynamic fields
        are borrowed from the corresponding plugin.
        """
        sharable_fields = getattr(self.plugin_instance, 'sharable_fields', [])
        form_mixin = self.plugin_instance().get_form(request)
        attrs = {name: field for name, field in form_mixin.base_fields.items() if name in sharable_fields}

        class Meta:
            untangled_fields = ['identifier']
            entangled_fields = {'glossary': list(attrs.keys())}

        attrs['Meta'] = Meta
        kwargs['form'] = type('SharedFieldsForm', (EntangledModelForm,), attrs)
        return super().get_form(request, obj, **kwargs)

    def has_add_permission(self, request):
        # always False, since a SharedGlossary can only be added by a plugin
        return False

    def add_view(self, request, form_url='', extra_context=None):
        raise AssertionError("This method shall never be called")

    def change_view(self, request, object_id, form_url='', extra_context=None):
        obj = self.get_object(request, object_id)
        self.plugin_instance = plugin_pool.get_plugin(obj.plugin_type)
        extra_context = dict(extra_context or {},
                             title=format_html(_("Change shared settings of '{}' plugin"), self.plugin_instance.name),
                             icon_fonts=IconFont.objects.all())
        return super().change_view(request, object_id, form_url, extra_context=extra_context)

    def used_by(self, obj):
        """
        Returns the number of plugins using this shared glossary
        """
        return SharableCascadeElement.objects.filter(shared_glossary=obj).count()
    used_by.short_description = _("Used by plugins")

    def render_change_form(self, request, context, add=False, change=False, form_url='', obj=None):
        try:
            context['media'] += self.plugin_instance().media
        except (AttributeError, KeyError):
            pass
        context.update(
            ring_plugin=self.plugin_instance().ring_plugin,
            ring_plugin_bases=dict((ring_plugin, ['django.cascade.{}'.format(b) for b in bases if b != 'SharableGlossaryMixin'])
                                   for ring_plugin, bases in CascadePluginMixinMetaclass.ring_plugin_bases.items())
        )
        return super().render_change_form(request, context, add, change, form_url, obj)

    def plugin_name(self, obj):
        plugin_instance = plugin_pool.get_plugin(obj.plugin_type)
        return plugin_instance.name
    plugin_name.short_description = _("Plugin Type")



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0020_page_icon_font'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='cascadepage',
            options={'verbose_name': 'Cascade Page Settings', 'verbose_name_plural': 'Cascade Page Settings'},
        ),
    ]



from django.core.exceptions import ImproperlyConfigured
from django.forms import MediaDefiningClass, ModelForm
from django.utils.functional import lazy
from django.utils.module_loading import import_string
from django.utils.text import format_lazy
from django.utils.safestring import SafeText, mark_safe
from entangled.forms import EntangledModelFormMixin
from cms.plugin_base import CMSPluginBaseMetaclass, CMSPluginBase
from cms.utils.compat.dj import is_installed
from cmsplugin_cascade import app_settings
from .mixins import CascadePluginMixin
from .models_base import CascadeModelBase
from .models import CascadeElement, SharableCascadeElement
from .generic.mixins import SectionMixin, SectionModelMixin
from .sharable.forms import SharableGlossaryMixin
from .strides import register_stride
from .extra_fields.mixins import ExtraFieldsMixin
from .hide_plugins import HidePluginMixin
from .render_template import RenderTemplateMixin
from .utils import remove_duplicates

mark_safe_lazy = lazy(mark_safe, str)

fake_proxy_models = {}


def create_proxy_model(name, model_mixins, base_model, attrs=None, module=None):
    """
    Create a Django Proxy Model on the fly, to be used by any Cascade Plugin.
    """
    from django.apps import apps

    class Meta:
        proxy = True
        app_label = 'cmsplugin_cascade'

    name = str(name + 'Model')
    try:
        Model = apps.get_registered_model(Meta.app_label, name)
    except LookupError:
        bases = model_mixins + (base_model,)
        attrs = dict(attrs or {}, Meta=Meta, __module__=module)
        Model = type(name, bases, attrs)
        fake_proxy_models[name] = bases
    return Model


class CascadePluginMixinMetaclass(MediaDefiningClass):
    ring_plugin_bases = {}

    def __new__(cls, name, bases, attrs):
        ring_plugin = attrs.get('ring_plugin')
        if ring_plugin:
            ring_plugin_bases = [b.ring_plugin for b in bases
                                 if hasattr(b, 'ring_plugin') and b.ring_plugin != ring_plugin]

            # remember the dependencies
            cls.ring_plugin_bases.setdefault(ring_plugin, [])
            cls.ring_plugin_bases[ring_plugin].extend(ring_plugin_bases)
            cls.ring_plugin_bases[ring_plugin] = remove_duplicates(cls.ring_plugin_bases[ring_plugin])

        new_class = super().__new__(cls, name, bases, attrs)
        return new_class


class CascadePluginMixinBase(metaclass=CascadePluginMixinMetaclass):
    """
    Use this as a base for mixin classes used by other CascadePlugins
    """


class CascadePluginBaseMetaclass(CascadePluginMixinMetaclass, CMSPluginBaseMetaclass):
    """
    All plugins from djangocms-cascade can be instantiated in different ways. In order to allow this
    by a user defined configuration, this meta-class conditionally inherits from additional mixin
    classes.
    """
    plugins_with_extra_fields = dict(app_settings.CMSPLUGIN_CASCADE['plugins_with_extra_fields'])
    plugins_with_extra_mixins = dict(app_settings.CMSPLUGIN_CASCADE['plugins_with_extra_mixins'])
    plugins_with_bookmark = list(app_settings.CMSPLUGIN_CASCADE['plugins_with_bookmark'])
    plugins_with_sharables = dict(app_settings.CMSPLUGIN_CASCADE['plugins_with_sharables'])
    plugins_with_extra_render_templates = app_settings.CMSPLUGIN_CASCADE['plugins_with_extra_render_templates'].keys()
    allow_plugin_hiding = app_settings.CMSPLUGIN_CASCADE['allow_plugin_hiding']
    exclude_hiding_plugin = list(app_settings.CMSPLUGIN_CASCADE['exclude_hiding_plugin'])

    def __new__(cls, name, bases, attrs):
        model_mixins = attrs.pop('model_mixins', ())
        if (cls.allow_plugin_hiding and name not in cls.exclude_hiding_plugin and 'name' in attrs and
            not attrs.get('text_enabled')):
            bases = (HidePluginMixin,) + bases
        if name in cls.plugins_with_extra_fields:
            bases = (ExtraFieldsMixin,) + bases
        if name in cls.plugins_with_extra_mixins:
            bases = (cls.plugins_with_extra_mixins[name],) + bases
        if name in cls.plugins_with_bookmark:
            bases = (SectionMixin,) + bases
            model_mixins = (SectionModelMixin,) + model_mixins
        if name in cls.plugins_with_sharables:
            bases = (SharableGlossaryMixin,) + bases
            attrs['sharable_fields'] = cls.plugins_with_sharables[name]
            base_model = SharableCascadeElement
        else:
            base_model = CascadeElement
        if name in cls.plugins_with_extra_render_templates:
            bases = (RenderTemplateMixin,) + bases
        if name == 'SegmentPlugin':
            # SegmentPlugin shall additionally inherit from configured mixin classes
            model_mixins += tuple(import_string(mc[0]) for mc in app_settings.CMSPLUGIN_CASCADE['segmentation_mixins'])
        if 'model' in attrs:
            # the plugin overrides the CascadeModel
            if not issubclass(attrs['model'], CascadeModelBase):
                msg = "Cascade Plugins, overriding the model, must inherit from `CascadeModelBase`."
                raise ImproperlyConfigured(msg)
        else:
            attrs['model'] = create_proxy_model(name, model_mixins, base_model, module=attrs.get('__module__'))
        if is_installed('reversion'):
            import reversion.revisions
            if not reversion.revisions.is_registered(base_model):
                reversion.revisions.register(base_model)
        # handle ambiguous plugin names by appending a symbol
        if 'name' in attrs and app_settings.CMSPLUGIN_CASCADE['plugin_prefix']:
            attrs['name'] = format_lazy('{}&nbsp;{}', app_settings.CMSPLUGIN_CASCADE['plugin_prefix'], attrs['name'])

        register_stride(name, bases, attrs, model_mixins)
        if name == 'CascadePluginBase':
            bases += (CascadePluginMixin, CMSPluginBase,)
        return super().__new__(cls, name, bases, attrs)


class TransparentWrapper:
    """
    Add this mixin class to other Cascade plugins, wishing to be added transparently between other
    plugins restricting parent-children relationships.
    For instance: A BootstrapColumnPlugin can only be added as a child to a RowPlugin. This means
    that no other wrapper can be added between those two plugins. By adding this mixin class we can
    allow any plugin to behave transparently, just as if it would not have be inserted into the DOM
    tree. When moving plugins in- and out of transparent wrapper plugins, always reload the page, so
    that the parent-children relationships can be updated.
    """
    child_plugins_cache = False
    parent_plugins_cache = False

    @classmethod
    def get_child_classes(cls, slot, page, instance=None):
        if hasattr(cls, 'direct_child_classes'):
            return cls.direct_child_classes
        child_classes = set(super().get_child_classes(slot, page, instance))
        while True:
            instance = instance.get_parent_instance() if instance and instance.parent else None
            if instance is None:
                child_classes.update(super().get_child_classes(slot, page, instance))
                return list(child_classes)
            if not issubclass(instance.plugin_class, TransparentWrapper):
                child_classes.update(instance.plugin_class.get_child_classes(slot, page, instance))
                return list(child_classes)

    @classmethod
    def get_parent_classes(cls, slot, page, instance=None):
        if hasattr(cls, 'direct_parent_classes'):
            return cls.direct_parent_classes
        parent_classes = set(super().get_parent_classes(slot, page, instance) or [])
        if isinstance(instance, CascadeElement):
            instance = instance.get_parent_instance() if instance and instance.parent else None
            if instance is not None:
                parent_classes.add(instance.plugin_type)
        return list(parent_classes)


class TransparentContainer(TransparentWrapper):
    """
    This mixin class marks each plugin inheriting from it, as a transparent container.
    Such a plugin is added to the global list of entitled parent plugins, which is required if we
    want to place and move all other Cascade plugins below this container.

    Often, transparent wrapping classes come in pairs. For instance the `AccordionPlugin` containing
    one or more `PanelPlugin`. Here the `AccordionPlugin` must inherit from `TransparentWrapper`,
    whereas the `AccordionPlugin` must inherit from the `TransparentContainer`.
    """
    @staticmethod
    def get_plugins():
        from cms.plugin_pool import plugin_pool
        global _leaf_transparent_plugins

        try:
            return _leaf_transparent_plugins
        except NameError:
            _leaf_transparent_plugins = [
                plugin.__name__ for plugin in plugin_pool.get_all_plugins()
                    if issubclass(plugin, TransparentContainer)
            ]
            return _leaf_transparent_plugins


class CascadeFormMixin(EntangledModelFormMixin):
    class Meta:
        entangled_fields = {'glossary': []}


class CascadePluginBase(metaclass=CascadePluginBaseMetaclass):
    change_form_template = 'cascade/admin/change_form.html'
    model_mixins = ()  # model mixins added to the final Django model
    parent_classes = None
    alien_child_classes = False
    form = CascadeFormMixin  # safety fallback for plugins without any form

    class Media:
        css = {'all': ['cascade/css/admin/partialfields.css', 'cascade/css/admin/editplugin.css']}
        js = ['cascade/js/underscore.js', 'cascade/js/ring.js']

    def __init__(self, model=None, admin_site=None):
        super().__init__(model, admin_site)

    def __repr__(self):
        return "<class '{}'>".format(self.__class__.__name__)

    @classmethod
    def super(cls, klass, instance):
        """
        Plugins inheriting from CascadePluginBaseMetaclass can have two different base classes,
        :class:`cmsplugin_cascade.plugin_base.CMSPluginBase` and :class:`cmsplugin_cascade.strides.StridePluginBase`.
        Therefore in order to call a method from an inherited class, use this "super" wrapping method.
        >>> cls.super(MyPlugin, self).a_method()
        """
        return super(klass, instance)

    @classmethod
    def _get_parent_classes_transparent(cls, slot, page, instance=None):
        """
        Return all parent classes including those marked as "transparent".
        """
        parent_classes = super().get_parent_classes(slot, page, instance)
        if parent_classes is None:
            if cls.get_require_parent(slot, page) is False:
                return
            parent_classes = []

        # add all plugins marked as 'transparent', since they all are potential parents
        parent_classes = set(parent_classes)
        parent_classes.update(TransparentContainer.get_plugins())
        return list(parent_classes)

    @classmethod
    def get_child_classes(cls, slot, page, instance=None):
        plugin_type = cls.__name__
        child_classes = set()
        for child_class in cls.get_child_plugin_candidates(slot, page):
            if issubclass(child_class, CascadePluginBase):
                own_child_classes = getattr(cls, 'child_classes', None) or []
                child_parent_classes = child_class._get_parent_classes_transparent(slot, page, instance)
                if isinstance(child_parent_classes, (list, tuple)) and plugin_type in child_parent_classes:
                    child_classes.add(child_class)
                elif plugin_type in own_child_classes:
                    child_classes.add(child_class)
                elif child_parent_classes is None:
                    child_classes.add(child_class)
            else:
                if cls.alien_child_classes and child_class.__name__ in app_settings.CMSPLUGIN_CASCADE['alien_plugins']:
                    child_classes.add(child_class)

        return list(cc.__name__ for cc in child_classes)

    @classmethod
    def get_parent_classes(cls, slot, page, instance=None):
        return cls._get_parent_classes_transparent(slot, page, instance)

    @classmethod
    def get_identifier(cls, instance):
        """
        Hook to return a description for the current model.
        """
        return SafeText()

    @classmethod
    def sanitize_model(cls, instance):
        """
        This method is called, before the model is written to the database. It can be overloaded
        to sanitize the current models, in case a parent model changed in a way, which might
        affect this plugin.
        This method shall return `True`, in case a model change was necessary, otherwise it shall
        return `False` to prevent a useless database update.
        """
        if instance.glossary is None:
            instance.glossary = {}
        return False

    @classmethod
    def get_data_representation(cls, instance):
        """
        Return a representation of the given instance suitable for a serialized representation.
        """
        return {'glossary': instance.glossary, 'pk': instance.pk}

    @classmethod
    def add_inline_elements(cls, instance, inlines):
        """
        Hook to create (sortable) inline elements for the given instance.
        """

    @classmethod
    def add_shared_reference(cls, instance, shared_glossary):
        """
        Hook to add a reference pointing onto an existing SharedGlossary instance.
        """

    def extend_children(self, parent, wanted_children, child_class, child_glossary=None):
        """
        Extend the number of children so that the parent object contains wanted children.
        No child will be removed if wanted_children is smaller than the current number of children.
        """
        from cms.api import add_plugin
        current_children = parent.get_num_children()
        for _ in range(current_children, wanted_children):
            child = add_plugin(parent.placeholder, child_class, parent.language, target=parent)
            if isinstance(child_glossary, dict):
                child.glossary.update(child_glossary)
            child.save()

    def get_form(self, request, obj=None, **kwargs):
        form = kwargs.get('form', self.form)
        assert issubclass(form, EntangledModelFormMixin), "Form must inherit from EntangledModelFormMixin"
        bases = (form,)
        if not issubclass(form, CascadeFormMixin):
            bases = (CascadeFormMixin,) + bases
        if not issubclass(form, ModelForm):
            bases += (ModelForm,)
        kwargs['form'] = type(form.__name__, bases, {})
        return super().get_form(request, obj, **kwargs)

    def get_parent_instance(self, request=None, obj=None):
        """
        Get the parent model instance corresponding to this plugin. When adding a new plugin, the
        parent might not be available. Therefore as fallback, pass in the request object.
        """
        try:
            parent_id = obj.parent_id
        except AttributeError:
            try:
                # TODO: self.parent presumably is not used anymore in CMS-3.4, because it doesn't
                # make sense anyway, since the plugin instances shall know their parents, not the
                # plugins.
                parent_id = self.parent.id
            except AttributeError:
                if request:
                    parent_id = request.GET.get('plugin_parent', None)
                    if parent_id is None:
                        from cms.models import CMSPlugin
                        try:
                            parent_id = CMSPlugin.objects.filter(id=request.resolver_match.args[0]
                                                                 ).only("parent_id").order_by('?').first().parent_id
                        except (AttributeError, IndexError):
                            parent_id = None
                else:
                    parent_id = None
        for model in CascadeModelBase._get_cascade_elements():
            try:
                return model.objects.get(id=parent_id)
            except model.DoesNotExist:
                continue

    def get_previous_instance(self, obj):
        """
        Return the previous plugin instance for the given object.
        This differs from `obj.get_prev_sibling()` which returns an unsorted sibling.
        """
        ordered_siblings = obj.get_siblings().filter(placeholder=obj.placeholder).order_by('position')
        pos = list(ordered_siblings).index(obj.cmsplugin_ptr)
        if pos > 0:
            prev_sibling = ordered_siblings[pos - 1]
            return prev_sibling.get_bound_plugin()

    def get_next_instance(self, obj):
        """
        Return the next plugin instance for the given object.
        This differs from `obj.get_next_sibling()` which returns an unsorted sibling.
        """
        ordered_siblings = obj.get_siblings().filter(placeholder=obj.placeholder).order_by('position')
        pos = list(ordered_siblings).index(obj.cmsplugin_ptr)
        if pos < ordered_siblings.count() - 1:
            next_sibling = ordered_siblings[pos + 1]
            return next_sibling.get_bound_plugin()

    def render_change_form(self, request, context, add=False, change=False, form_url='', obj=None):
        ring_plugin_bases = {ring_plugin: ['django.cascade.{}'.format(b) for b in bases]
                             for ring_plugin, bases in CascadePluginMixinMetaclass.ring_plugin_bases.items()}
        context.update(
            ring_plugin_bases=ring_plugin_bases,
            plugin_title=format_lazy("{} {} Plugin", self.module, self.name),
            plugin_intro=mark_safe(getattr(self, 'intro_html', '')),
            plugin_footnote=mark_safe(getattr(self, 'footnote_html', '')),
        )
        if hasattr(self, 'ring_plugin'):
            context.update(
                ring_plugin=self.ring_plugin,
            )
        context['empty_form'] = not context['adminform'].form._meta.entangled_fields.get('glossary')
        return super().render_change_form(request, context, add, change, form_url, obj)

    def in_edit_mode(self, request, placeholder):
        """
        Returns True, if the plugin is in "edit mode".
        """
        toolbar = getattr(request, 'toolbar', None)
        edit_mode = getattr(toolbar, 'edit_mode_active', False) and getattr(placeholder, 'is_editable', True)
        if edit_mode:
            edit_mode = placeholder.has_change_permission(request.user)
        return edit_mode



from django.core.exceptions import ImproperlyConfigured
from django.forms.fields import BooleanField
from django.utils.translation import gettext_lazy as _
from django.template import engines

from entangled.forms import EntangledModelFormMixin


class HidePluginFormMixin(EntangledModelFormMixin):
    hide_plugin = BooleanField(
        label=_("Hide plugin"),
        required=False,
        help_text=_("Hide this plugin and all of it's children.")
    )

    class Meta:
        entangled_fields = {'glossary': ['hide_plugin']}


class HidePluginMixin:
    """
    This mixin class adds a checkbox to each named plugin, which if checked hides that
    plugin during the rendering phase.
    """
    suppress_template = engines['django'].from_string('')
    hiding_template_string = '''
{{% load cms_tags %}}
<div style="display: none;">
{{% for plugin in instance.child_plugin_instances %}}{{% render_plugin plugin %}}{{% endfor %}}
</div>
<style>
div.cms .cms-structure .cms-draggable-{plugin_id} .cms-dragitem {{
color: gray;
background-color: lightgray;
background-image: repeating-linear-gradient(-45deg, transparent, transparent 4px, rgba(255,255,255,.5) 4px, rgba(255,255,255,.5) 8px);
background-size: contain;
}}
</style>
'''

    def get_form(self, request, obj=None, **kwargs):
        form = kwargs.get('form', self.form)
        assert issubclass(form, EntangledModelFormMixin), "Form must inherit from EntangledModelFormMixin"
        kwargs['form'] = type(form.__name__, (HidePluginFormMixin, form), {})
        return super().get_form(request, obj, **kwargs)

    def get_render_template(self, context, instance, placeholder):
        if instance.glossary.get('hide_plugin'):
            if self.in_edit_mode(context['request'], placeholder):
                # in edit mode we actually must render the children, otherwise they won't show
                # up in Structure Mode
                template_string = self.hiding_template_string.format(plugin_id=instance.pk)
                return engines['django'].from_string(template_string)
            else:
                return self.suppress_template

        super_self = super(HidePluginMixin, self)
        if hasattr(super_self, 'get_render_template'):
            template = super_self.get_render_template(context, instance, placeholder)
        elif getattr(self, 'render_template', False):
            template = getattr(self, 'render_template', False)
        else:
            template = None
        if not template:
            raise ImproperlyConfigured("Plugin {} has no render_template.".format(self.__class__))
        return template



from django.utils.safestring import mark_safe
from cmsplugin_cascade.models import IconFont
from cmsplugin_cascade.plugin_base import CascadePluginMixinBase
from entangled.utils import get_related_object


class IconPluginMixin(CascadePluginMixinBase):
    change_form_template = 'cascade/admin/change_form.html'
    ring_plugin = 'IconPluginMixin'

    class Media:
        css = {'all': ['cascade/css/admin/iconplugin.css']}
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/iconpluginmixin.js']

    def changeform_view(self, request, object_id=None, form_url='', extra_context=None):
        extra_context = dict(extra_context or {}, icon_fonts=IconFont.objects.all())
        return super().changeform_view(request, object_id, form_url, extra_context)

    def render(self, context, instance, placeholder):
        context = super().render(context, instance, placeholder)
        icon_font = get_related_object(instance.glossary, 'icon_font')
        symbol = instance.glossary.get('symbol')
        if icon_font and symbol:
            prefix = icon_font.config_data.get('css_prefix_text', 'icon-')
            context.update({
                'stylesheet_url': icon_font.get_stylesheet_url(),
                'icon_font_class': mark_safe('{}{}'.format(prefix, symbol)),
            })
        return context



from django.db import migrations

def forwards(apps, schema_editor):
    CascadeElement = apps.get_model('cmsplugin_cascade', 'CascadeElement')
    for element in CascadeElement.objects.using(schema_editor.connection.alias).filter(plugin_type='HeadingPlugin'):
        head_size = element.glossary.pop('head_size', None)
        if head_size:
            element.glossary['tag_type'] = 'h{}'.format(head_size)
            element.save()


def backwards(apps, schema_editor):
    CascadeElement = apps.get_model('cmsplugin_cascade', 'CascadeElement')
    for element in CascadeElement.objects.using(schema_editor.connection.alias).filter(plugin_type='HeadingPlugin'):
        tag_type = element.glossary.pop('tag_type', None)
        if tag_type and len(tag_type) == 2:
            element.glossary['head_size'] = tag_type[1]
            element.save()


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0009_cascadepage'),
    ]

    operations = [
        migrations.RunPython(forwards, reverse_code=backwards),
    ]



from django.forms.fields import CharField
from django.utils.translation import gettext_lazy as _

from entangled.forms import EntangledModelFormMixin, EntangledField
from entangled.utils import get_related_object
from cmsplugin_cascade.fields import CascadeImageField


class ImageFormMixin(EntangledModelFormMixin):
    image_file = CascadeImageField()

    image_title = CharField(
        label=_('Image Title'),
        required=False,
        help_text=_("Caption text added to the 'title' attribute of the <img> element."),
    )

    alt_tag = CharField(
        label=_('Alternative Description'),
        required=False,
        help_text=_("Textual description of the image added to the 'alt' tag of the <img> element."),
    )

    _image_properties = EntangledField()

    class Meta:
        entangled_fields = {'glossary': ['image_file', 'image_title', 'alt_tag', '_image_properties']}

    def __init__(self, *args, **kwargs):
        if not getattr(self, 'require_image', True):
            self.base_fields['image_file'].required = False
        super().__init__(*args, **kwargs)

    def clean_image_file(self):
        image_file = self.cleaned_data['image_file']
        # _image_properties are just a cached representation, maybe useless
        if image_file:
            self.cleaned_data['_image_properties'] = {
                'width': image_file._width,
                'height': image_file._height,
                'exif_orientation': image_file.exif.get('Orientation', 1),
            }
        return image_file


class ImagePropertyMixin:
    """
    A mixin class to convert a CascadeElement into a proxy model for rendering an image element.
    """
    def __str__(self):
        try:
            return self.plugin_class.get_identifier(self)
        except AttributeError:
            return str(self.image)

    @property
    def image(self):
        if not hasattr(self, '_image_file'):
            self._image_file = get_related_object(self.glossary, 'image_file')
        return self._image_file

    def post_copy(self, old_instance, new_old_ziplist):
        # by saving this model after the full tree has been copied, ``<Any>ImagePlugin.sanitize_model()``
        # is invoked a second time with the now complete information of all column siblings.
        self.save(sanitize_only=True)



from django.contrib import admin
from django.contrib import messages
from django.forms import widgets
from django.forms.utils import flatatt
from django.db.models import JSONField
from django.templatetags.static import static
from django.utils.html import format_html
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from cms.models.placeholderpluginmodel import PlaceholderReference
from cmsplugin_cascade.clipboard.utils import deserialize_to_clipboard, serialize_from_placeholder
from cmsplugin_cascade.models import CascadeClipboard


class JSONAdminWidget(widgets.Textarea):
    def __init__(self):
        attrs = {'cols': '40', 'rows': '3'}
        super(JSONAdminWidget, self).__init__(attrs)

    def render(self, name, value, attrs=None, renderer=None):
        if value is None:
            value = ''
        final_attrs = self.build_attrs(self.attrs, extra_attrs=dict(attrs, name=name))
        id_data = attrs.get('id', 'id_data')
        clippy_url = static('cascade/admin/clippy.svg')
        return format_html('<textarea{0}>\r\n{1}</textarea> '
            '<button data-clipboard-target="#{2}" type="button" title="{4}" class="clip-btn">'
                '<img src="{3}" alt="{4}">'
            '</button>\n'
            '<div class="status-line"><label></label><strong id="pasted_success">{5}</strong>'
            '<strong id="copied_success">{6}</strong></div>',
            flatatt(final_attrs), str(value), id_data, clippy_url,
            _("Copy to Clipboard"),
            _("Successfully pasted JSON data"),
            _("Successfully copied JSON data"))


@admin.register(CascadeClipboard)
class CascadeClipboardAdmin(admin.ModelAdmin):
    fields = ['identifier', ('created_by', 'created_at', 'last_accessed_at'), 'save_clipboard', 'restore_clipboard', 'data']
    readonly_fields = ['created_by', 'created_at', 'last_accessed_at', 'save_clipboard', 'restore_clipboard']
    formfield_overrides = {
        JSONField: {'widget': JSONAdminWidget},
    }
    list_display = ['identifier', 'created_by', 'created_at']

    class Media:
        css = {'all': ['cascade/css/admin/clipboard.css']}
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/clipboard.js']

    def save_clipboard(self, obj):
        return format_html('<input type="submit" value="{}" class="default pull-left" name="save_clipboard" />',
                           _("Insert Data"))
    save_clipboard.short_description = _("From CMS Clipboard")

    def restore_clipboard(self, obj):
        return format_html('<input type="submit" value="{}" class="default pull-left" name="restore_clipboard" />',
                           _("Restore Data"))
    restore_clipboard.short_description = _("To CMS Clipboard")

    def save_model(self, request, obj, form, change):
        if request.POST.get('save_clipboard'):
            placeholder_reference = PlaceholderReference.objects.last()
            if placeholder_reference:
                placeholder = placeholder_reference.placeholder_ref
                obj.data = serialize_from_placeholder(placeholder, self.admin_site)
            request.POST = request.POST.copy()
            request.POST['_continue'] = True
            messages.add_message(request, messages.INFO, _("The CMS clipboard has been persisted into the database."))
        if request.POST.get('restore_clipboard'):
            request.POST = request.POST.copy()
            request.POST['_continue'] = True
            messages.add_message(request, messages.INFO, _("Persisted content has been restored to CMS clipboard."))
        if request.POST.get('restore_clipboard'):
            deserialize_to_clipboard(request, obj.data)
            obj.last_accessed_at = now()
        super().save_model(request, obj, form, change)




from django.db import migrations, models
import django.db.models.deletion


def forwards(apps, schema_editor):
    print("Skipping this migration.")


def backwards(apps, schema_editor):
    print("Backward migration not implemented.")


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0019_verbose_table_names'),
    ]

    operations = [
        migrations.AddField(
            model_name='cascadepage',
            name='icon_font',
            field=models.ForeignKey(blank=True, help_text='Set Icon Font globally for this page', null=True, on_delete=django.db.models.deletion.CASCADE, to='cmsplugin_cascade.IconFont', verbose_name='Icon Font'),
        ),
        migrations.RunPython(forwards, reverse_code=backwards),
    ]



from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from cms.models import Placeholder
from cmsplugin_cascade.models import CascadeClipboard


class ClipboardBaseForm(forms.Form):
    placeholder = forms.ModelChoiceField(
        queryset=Placeholder.objects.all(),
        required=True,
        widget=forms.HiddenInput(),
    )

    language = forms.ChoiceField(
        choices=settings.LANGUAGES,
        required=True,
        widget=forms.HiddenInput(),
    )

    def clean_identifier(self):
        identifier = self.cleaned_data['identifier']
        if CascadeClipboard.objects.filter(identifier=identifier).exists():
            msg = _("This identifier has already been used, please choose another one.")
            raise ValidationError(msg)
        return identifier



from copy import copy
from enum import Enum, unique
from functools import reduce
import itertools
from operator import add
import re

from django.utils.translation import gettext_lazy as _


class BootstrapException(Exception):
    """
    Raised if arrangement of Bootstrap-4 elements would render weird results.
    """


@unique
class Breakpoint(Enum):
    """
    Enumerate the five breakpoints defined by the Bootstrap-4 CSS framework.
    """
    xs = 0
    sm = 1
    md = 2
    lg = 3
    xl = 4

    @classmethod
    def range(cls, first, last):
        """
        Iterate over all elements, starting from `first` until `last`.
        The last element is included.
        """
        if first: first = first.value
        if last: last = last.value + 1
        return itertools.islice(cls, first, last)

    def __gt__(self, other):
        return self.value > other.value

    def __ge__(self, other):
        return self.value >= other.value

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __iter__(self):
        yield self.xs
        yield self.sm
        yield self.md
        yield self.lg
        yield self.xl

    @property
    def label(self):
        return [
            _("Portrait Phones"),
            _("Landscape Phones"),
            _("Tablets"),
            _("Laptops"),
            _("Large Desktops"),
        ][self.value]

    @property
    def media_query(self):
        return [
            '(max-width: 575.98px)',
            '(min-width: 576px) and (max-width: 767.98px)',
            '(min-width: 768px) and (max-width: 991.98px)',
            '(min-width: 992px) and (max-width: 1199.98px)',
            '(min-width: 1200px)',
        ][self.value]


class Bound:
    def __init__(self, min, max):
        self.min = float(min)
        self.max = float(max)

    def __copy__(self):
        return type(self)(self.min, self.max)

    def __eq__(self, other):
        return round(self.min, 1) == round(other.min, 1) and round(self.max, 1) == round(other.max, 1)

    def __add__(self, other):
        return Bound(
            self.min + other.min,
            self.max + other.max,
        )

    def __sub__(self, other):
        return Bound(
            self.min - other.min,
            self.max - other.max,
        )

    def __repr__(self):
        return "{}(min={}, max={})".format(self.__class__.__name__, self.min, self.max)

    def extend(self, other):
        self.min = min(self.min, other.min)
        self.max = max(self.max, other.max)


default_bounds = {
    Breakpoint.xs: Bound(320, 572),
    Breakpoint.sm: Bound(540, 540),
    Breakpoint.md: Bound(720, 720),
    Breakpoint.lg: Bound(960, 960),
    Breakpoint.xl: Bound(1140, 1140),
}

fluid_bounds = {
    Breakpoint.xs: Bound(320, 576),
    Breakpoint.sm: Bound(576, 768),
    Breakpoint.md: Bound(768, 992),
    Breakpoint.lg: Bound(992, 1200),
    Breakpoint.xl: Bound(1200, 1980),
}


class Break:
    def __init__(self, breakpoint, classes, narrower=None):
        self.breakpoint = breakpoint
        self.fixed_units = 0
        self.flex_column = False
        self.auto_column = False
        self._normalize_col_classes(classes)
        if isinstance(narrower, Break):
            self._inherit_from(narrower)
        self.bound = None

    def _normalize_col_classes(self, classes):
        if self.breakpoint == Breakpoint.xs:
            fixed_pat = re.compile(r'^col-(\d+)$')
            flex_pat = re.compile(r'^col$')
            auto_pat = re.compile(r'^col-auto$')
        else:
            fixed_pat = re.compile(r'^col-{}-(\d+)$'.format(self.breakpoint.name))
            flex_pat = re.compile(r'^col-{}$'.format(self.breakpoint.name))
            auto_pat = re.compile(r'^col-{}-auto$'.format(self.breakpoint.name))
        for col_class in classes:
            # look for CSS classes matching fixed size columns
            fixed = fixed_pat.match(col_class)
            if fixed:
                if self.fixed_units or self.flex_column or self.auto_column:
                    raise BootstrapException("Can not mix fixed- with flex- or auto-column")
                units = int(fixed.group(1))
                if units < 1 or units > 12:
                    raise BootstrapException("Column units value {} out of range".format(units))
                self.fixed_units = units

            # look for CSS classes matching flex columns
            flex = flex_pat.match(col_class)
            if flex:
                if self.fixed_units or self.flex_column or self.auto_column:
                    raise BootstrapException("Can not mix flex- with fixed- or auto-column")
                self.flex_column = True

            # look for CSS classes matching auto columns
            auto = auto_pat.match(col_class)
            if auto:
                if self.fixed_units or self.flex_column or self.auto_column:
                    raise BootstrapException("Can not mix auto- with fixed- or flex-column")
                self.auto_column = True

    def _inherit_from(self, narrower):
        if self.breakpoint <= narrower.breakpoint:
            raise BootstrapException("Can only inherit column bounds from narrower breakpoint")
        if self.fixed_units == 0 and self.flex_column is False and self.auto_column is False:
            self.fixed_units = narrower.fixed_units
            self.flex_column = narrower.flex_column
            self.auto_column = narrower.auto_column

    def __copy__(self):
        newone = type(self)()
        if self.bound:
            newone.bound = dict(self.bound)
        return newone

    def __repr__(self):
        return "{}({}: fixed={}, flex={}, auto={})".format(
            self.__class__.__name__, self.breakpoint.name, self.fixed_units, self.flex_column, self.auto_column)


class Bootstrap4Container(list):
    """
    Abstracts a Bootstrap-4 container element, such as ``<div class="container">...</div>``, so that the minimum and
    maximum widths each each child row can be computed.
    Each container object is a list of one to many ``Bootstrap4Row`` instances.
    In order to model a "fluid" container, use ``fluid_bounds`` during construction.
    """
    # def __init__(self, bounds=app_settings.CMSPLUGIN_CASCADE['bootstrap4']['default_bounds']):
    def __init__(self, bounds=default_bounds):
            self.bounds = bounds

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, ', '.join([repr(o) for o in self]))

    def add_row(self, row):
        if isinstance(row.parent, (Bootstrap4Container, Bootstrap4Column)):
            # detach from previous container or column
            pos = row.parent.index(row)
            row.parent.pop(pos)
        row.parent = self
        row.bounds = dict(self.bounds)
        self.append(row)
        return row


class Bootstrap4Row(list):
    """
    Abstracts a Bootstrap-4 row element, such as ``<div class="row">...</div>``, so that the minimum and maximum widths
    each each child columns can be computed.
    Each row object is a list of 1 to many ``Bootstrap4Column`` instances.
    """
    parent = None
    bounds = None

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, ', '.join([repr(o) for o in self]))

    def add_column(self, column):
        if isinstance(column.parent, Bootstrap4Row):
            pos = column.parent.index(column)
            column.parent.pop(pos)
        column.parent = self
        self.append(column)

    def compute_column_bounds(self):
        assert isinstance(self.bounds, dict), "Expected `bounds` to be a dict."
        for bp in [Breakpoint.xs, Breakpoint.sm, Breakpoint.md, Breakpoint.lg, Breakpoint.xl]:
            if bp in self.bounds:
                remaining_width = copy(self.bounds[bp])

            # first compute the bounds of columns with a fixed width
            for column in self:
                if column.breaks[bp].fixed_units:
                    assert column.breaks[bp].bound is None, "Expected `column.breaks[bp].bound` to be None."
                    column.breaks[bp].bound = Bound(
                        column.breaks[bp].fixed_units * self.bounds[bp].min / 12,
                        column.breaks[bp].fixed_units * self.bounds[bp].max / 12,
                    )
                    remaining_width -= column.breaks[bp].bound

            flex_columns = reduce(add, [int(col.breaks[bp].flex_column) for col in self], 0)
            auto_columns = reduce(add, [int(col.breaks[bp].auto_column) for col in self], 0)
            if auto_columns:
                # we use auto-columns, therefore estimate the min- and max values
                for column in self:
                    if column.breaks[bp].flex_column or column.breaks[bp].auto_column:
                        assert column.breaks[bp].bound is None, "Expected `column.breaks[bp].bound` to be None."
                        column.breaks[bp].bound = Bound(
                            30,
                            remaining_width.max - 30 * (flex_columns + auto_columns),
                        )
            else:
                # we use flex-columns exclusively, therefore subdivide the remaining width
                for column in self:
                    if column.breaks[bp].flex_column:
                        assert column.breaks[bp].bound is None, "Expected `column.breaks[bp].bound` to be None."
                        column.breaks[bp].bound = Bound(
                            remaining_width.min / flex_columns,
                            remaining_width.max / flex_columns,
                        )


class Bootstrap4Column(list):
    """
    Abstracts a Bootstrap-4 column element, such as ``<div class="col...">...</div>``, which shall be added to a
    ``Bootstrap4Row`` element.
    Each column may contain elements of type ``Bootstrap4Row``.
    For each breakpoint, a columns knows its extensions.
    """
    parent = None

    def __init__(self, classes=[]):
        if isinstance(classes, str):
            classes = classes.split()
        narrower = None
        self.breaks = {}
        for bp in Breakpoint:
            self.breaks[bp] = Break(bp, classes, narrower)
            narrower = self.breaks[bp]

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, ', '.join([repr(self.breaks[bp]) for bp in Breakpoint]))

    def __copy__(self):
        newone = type(self)()
        newone.__dict__.update(self.__dict__)
        return newone

    def add_row(self, row):
        if isinstance(row.parent, (Bootstrap4Container, Bootstrap4Column)):
            # detach from previous container or column
            pos = row.parent.index(row)
            row.parent.pop(pos)
            row.parent.bounds = None
        row.parent = self
        row.bounds = dict((bp, self.get_bound(bp)) for bp in Breakpoint)
        self.append(row)
        return row

    def get_bound(self, breakpoint):
        if self.breaks[breakpoint].bound is None:
            self.parent.compute_column_bounds()
            assert self.breaks[breakpoint].bound, "Expected `column.breaks[bp].bound` not to be None."
        return self.breaks[breakpoint].bound

    def get_min_max_bounds(self):
        """
        Return a dict of min- and max-values for the given column.
        This is required to estimate the bounds of images.
        """
        bound = Bound(999999.0, 0.0)
        for bp in Breakpoint:
            bound.extend(self.get_bound(bp))
        return {'min': bound.min, 'max': bound.max}



import json

from django.contrib.admin import site as default_admin_site
from django.contrib.admin.helpers import AdminForm
from django.core.exceptions import PermissionDenied
from django.forms import CharField, ModelChoiceField
from django.shortcuts import render
from django.template.response import TemplateResponse
from django.urls import re_path, reverse
from django.utils.http import urlencode
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from cms.plugin_base import CMSPluginBase, PluginMenuItem
from cms.plugin_pool import plugin_pool
from cms.toolbar.utils import get_plugin_tree_as_json
from cms.utils import get_language_from_request
from cmsplugin_cascade.clipboard.forms import ClipboardBaseForm
from cmsplugin_cascade.clipboard.utils import deserialize_to_clipboard, serialize_from_placeholder
from cmsplugin_cascade.models import CascadeClipboard


class CascadeClipboardPlugin(CMSPluginBase):
    render_plugin = False
    change_form_template = 'admin/cms/page/plugin/change_form.html'

    def get_plugin_urls(self):
        urlpatterns = [
            re_path(r'^export-plugins/$', self.export_plugins_view, name='export_clipboard_plugins'),
            re_path(r'^import-plugins/$', self.import_plugins_view, name='import_clipboard_plugins'),
        ]
        return urlpatterns

    @classmethod
    def get_extra_placeholder_menu_items(cls, request, placeholder):
        data = urlencode({
            'placeholder': placeholder.pk,
            'language': get_language_from_request(request),
        })
        return [
            PluginMenuItem(
                _("Export to Clipboard"),
                reverse('admin:export_clipboard_plugins') + '?' + data,
                data={},
                action='modal',
                attributes={
                    'icon': 'export',
                },
            ),
            PluginMenuItem(
                _("Import from Clipboard"),
                reverse('admin:import_clipboard_plugins') + '?' + data,
                data={},
                action='modal',
                attributes={
                    'icon': 'import',
                },
            )
        ]

    def render_modal_window(self, request, form):
        """
        Render a modal popup window with a select box to edit the form
        """
        opts = self.model._meta
        fieldsets = [(None, {'fields': list(form.fields)})]
        adminForm = AdminForm(form, fieldsets, {}, [])
        context = {
            **default_admin_site.each_context(request),
            'title': form.title,
            'adminform': adminForm,
            'add': False,
            'change': True,
            'save_as': False,
            'has_add_permission': False,
            'has_change_permission': True,
            'opts': opts,
            'root_path': reverse('admin:index'),
            'is_popup': True,
            'app_label': opts.app_label,
            'media': self.media + form.media,
        }
        return TemplateResponse(request, self.change_form_template, context)

    def import_plugins_view(self, request):
        # TODO: check for permissions

        title = _("Import from Clipboard")
        if request.method == 'GET':
            Form = type('ClipboardImportForm', (ClipboardBaseForm,), {
                'clipboard': ModelChoiceField(
                    queryset=CascadeClipboard.objects.all(),
                    label=_("Select Clipboard Content"),
                    required=False,
                ),
                'title': title,
            })
            form = Form(request.GET)
            assert form.is_valid()
        elif request.method == 'POST':
            Form = type('ClipboardImportForm', (ClipboardBaseForm,), {
                'clipboard': ModelChoiceField(
                    queryset=CascadeClipboard.objects.all(),
                    label=_("Select Clipboard Content"),
                ),
                'title': title,
            })
            form = Form(request.POST)
            if form.is_valid():
                return self.paste_from_clipboard(request, form)
        return self.render_modal_window(request, form)

    def paste_from_clipboard(self, request, form):
        placeholder = form.cleaned_data['placeholder']
        language = form.cleaned_data['language']
        cascade_clipboard = form.cleaned_data['clipboard']
        tree_order = placeholder.get_plugin_tree_order(language)
        deserialize_to_clipboard(request, cascade_clipboard.data)
        cascade_clipboard.last_accessed_at = now()
        cascade_clipboard.save(update_fields=['last_accessed_at'])

        # detach plugins from clipboard and reattach them to current placeholder
        cb_placeholder_plugin = request.toolbar.clipboard.cmsplugin_set.filter(plugin_type='PlaceholderPlugin').first()
        cb_placeholder_instance, _ = cb_placeholder_plugin.get_plugin_instance()
        new_plugins = cb_placeholder_instance.placeholder_ref.get_plugins()
        new_plugins.update(placeholder=placeholder)

        # reorder root plugins in placeholder
        root_plugins = placeholder.get_plugins(language).filter(parent__isnull=True).order_by('changed_date')
        for position, plugin in enumerate(root_plugins.iterator()):
            plugin.update(position=position)
        placeholder.mark_as_dirty(language, clear_cache=False)

        # create a list of pasted plugins to be added to the structure view
        all_plugins = placeholder.get_plugins(language)
        if all_plugins.exists():
            new_plugins = placeholder.get_plugins(language).exclude(pk__in=tree_order)
            data = json.loads(get_plugin_tree_as_json(request, list(new_plugins)))
            data['plugin_order'] = tree_order + ['__COPY__']
        else:
            return render(request, 'cascade/admin/clipboard_reload_page.html')
        data['target_placeholder_id'] = placeholder.pk
        context = {'structure_data': json.dumps(data)}
        return render(request, 'cascade/admin/clipboard_paste_plugins.html', context)

    def export_plugins_view(self, request):
        if not request.user.is_staff:
            raise PermissionDenied

        title = _("Export to Clipboard")
        if request.method == 'GET':
            Form = type('ClipboardExportForm', (ClipboardBaseForm,), {
                'identifier': CharField(required=False),
                'title': title,
            })
            form = Form(request.GET)
            assert form.is_valid()
        elif request.method == 'POST':
            Form = type('ClipboardExportForm', (ClipboardBaseForm,), {
                'identifier': CharField(),
                'title': title,
            })
            form = Form(request.POST)
            if form.is_valid():
                return self.add_to_clipboard(request, form)
        return self.render_modal_window(request, form)

    def add_to_clipboard(self, request, form):
        placeholder = form.cleaned_data['placeholder']
        language = form.cleaned_data['language']
        identifier = form.cleaned_data['identifier']
        data = serialize_from_placeholder(placeholder)
        CascadeClipboard.objects.create(
            identifier=identifier,
            data=data,
            created_by=request.user,
        )
        return render(request, 'cascade/admin/clipboard_close_frame.html', {})

plugin_pool.register_plugin(CascadeClipboardPlugin)



from django.forms.fields import ChoiceField, IntegerField
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin
from cms.plugin_pool import plugin_pool
from cms.models.pagemodel import Page
from .plugin_base import BootstrapPluginBase


class SecondaryMenuFormMixin(EntangledModelFormMixin):
    page_id = ChoiceField(
        label=_("CMS Page Id"),
        help_text = _("Select a CMS page with a given unique Id (in advanced settings)."),
    )

    offset = IntegerField(
        label=_("Offset"),
        initial=0,
        help_text=_("Starting from which child menu."),
    )

    limit = IntegerField(
        label=_("Limit"),
        initial=100,
        help_text=_("Number of child menus."),
    )

    class Meta:
        entangled_fields = {'glossary': ['page_id', 'offset', 'limit']}

    def __init__(self, *args, **kwargs):
        choices = [(p.reverse_id, str(p)) for p in Page.objects.filter(reverse_id__isnull=False, publisher_is_draft=False)]
        self.base_fields['page_id'].choices = choices
        super().__init__(*args, **kwargs)


class BootstrapSecondaryMenuPlugin(BootstrapPluginBase):
    """
    Use this plugin to display a secondary menu in arbitrary locations.
    This renders links onto  all CMS pages, which are children of the selected Page Id.
    """
    name = _("Secondary Menu")
    default_css_class = 'list-group'
    require_parent = False
    parent_classes = None
    allow_children = False
    form = SecondaryMenuFormMixin
    render_template = 'cascade/bootstrap4/secmenu-list-group.html'

    @classmethod
    def get_identifier(cls, obj):
        return mark_safe(obj.glossary.get('page_id', ''))

    def render(self, context, instance, placeholder):
        context = self.super(BootstrapSecondaryMenuPlugin, self).render(context, instance, placeholder)
        context.update({
            'page_id': instance.glossary['page_id'],
            'offset': instance.glossary.get('offset', 0),
            'limit': instance.glossary.get('limit', 100),
        })
        return context

    @classmethod
    def sanitize_model(cls, instance):
        try:
            if int(instance.glossary['offset']) < 0 or int(instance.glossary['limit']) < 0:
                raise ValueError()
        except (KeyError, ValueError):
            instance.glossary['offset'] = 0
            instance.glossary['limit'] = 100

plugin_pool.register_plugin(BootstrapSecondaryMenuPlugin)



from django.forms import widgets
from django.forms.fields import BooleanField, CharField, ChoiceField, MultipleChoiceField
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.icon.plugin_base import IconPluginMixin
from cmsplugin_cascade.icon.forms import IconFormMixin
from cmsplugin_cascade.link.config import LinkPluginBase, LinkFormMixin
from cmsplugin_cascade.link.plugin_base import LinkElementMixin


class ButtonTypeWidget(widgets.RadioSelect):
    """
    Render sample buttons in different colors in the button's backend editor.
    """
    template_name = 'cascade/admin/widgets/button_types.html'


class ButtonSizeWidget(widgets.RadioSelect):
    """
    Render sample buttons in different sizes in the button's backend editor.
    """
    template_name = 'cascade/admin/widgets/button_sizes.html'


class ButtonFormMixin(EntangledModelFormMixin):
    BUTTON_TYPES = [
        ('btn-primary', _("Primary")),
        ('btn-secondary', _("Secondary")),
        ('btn-success', _("Success")),
        ('btn-danger', _("Danger")),
        ('btn-warning', _("Warning")),
        ('btn-info', _("Info")),
        ('btn-light', _("Light")),
        ('btn-dark', _("Dark")),
        ('btn-link', _("Link")),
        ('btn-outline-primary', _("Primary")),
        ('btn-outline-secondary', _("Secondary")),
        ('btn-outline-success', _("Success")),
        ('btn-outline-danger', _("Danger")),
        ('btn-outline-warning', _("Warning")),
        ('btn-outline-info', _("Info")),
        ('btn-outline-light', _("Light")),
        ('btn-outline-dark', _("Dark")),
        ('btn-outline-link', _("Link")),
    ]

    BUTTON_SIZES = [
        ('btn-lg', _("Large button")),
        ('', _("Default button")),
        ('btn-sm', _("Small button")),
    ]

    link_content = CharField(
        required=False,
        label=_("Button Content"),
        widget=widgets.TextInput(attrs={'size': 50}),
    )

    button_type = ChoiceField(
        label=_("Button Type"),
        widget=ButtonTypeWidget(choices=BUTTON_TYPES),
        choices=BUTTON_TYPES,
        initial='btn-primary',
        help_text=_("Display Link using this Button Style")
    )

    button_size = ChoiceField(
        label=_("Button Size"),
        widget=ButtonSizeWidget(choices=BUTTON_SIZES),
        choices=BUTTON_SIZES,
        initial='',
        required=False,
        help_text=_("Display Link using this Button Size")
    )

    button_options = MultipleChoiceField(
        label=_("Button Options"),
        choices=[
            ('btn-block', _('Block level')),
            ('disabled', _('Disabled')),
        ],
        required=False,
        widget=widgets.CheckboxSelectMultiple,
    )

    stretched_link = BooleanField(
        label=_("Stretched link"),
        required=False,
        help_text=_("Stretched-link utility to make any anchor the size of it’s nearest position: " \
                    "relative parent, perfect for entirely clickable cards!")
    )

    icon_align = ChoiceField(
        label=_("Icon alignment"),
        choices=[
            ('icon-left', _("Icon placed left")),
            ('icon-right', _("Icon placed right")),
        ],
        widget=widgets.RadioSelect,
        initial='icon-right',
        help_text=_("Add an Icon before or after the button content."),
    )

    class Meta:
        entangled_fields = {'glossary': ['link_content', 'button_type', 'button_size', 'button_options', 'icon_align',
                                         'stretched_link']}


class BootstrapButtonMixin(IconPluginMixin):
    require_parent = True
    parent_classes = ['BootstrapColumnPlugin', 'SimpleWrapperPlugin']
    render_template = 'cascade/bootstrap4/button.html'
    allow_children = False
    default_css_class = 'btn'
    default_css_attributes = ['button_type', 'button_size', 'button_options', 'stretched_link']
    ring_plugin = 'ButtonMixin'

    class Media:
        css = {'all': ['cascade/css/admin/bootstrap4-buttons.css', 'cascade/css/admin/iconplugin.css']}
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/buttonmixin.js']

    def render(self, context, instance, placeholder):
        context = super().render(context, instance, placeholder)
        if 'icon_font_class' in context:
            mini_template = '{0}<i class="{1} {2}" aria-hidden="true"></i>{3}'
            icon_align = instance.glossary.get('icon_align')
            if icon_align == 'icon-left':
                context['icon_left'] = format_html(mini_template, '', context['icon_font_class'], 'cascade-icon-left',
                                                   ' ')
            elif icon_align == 'icon-right':
                context['icon_right'] = format_html(mini_template, ' ', context['icon_font_class'],
                                                    'cascade-icon-right', '')
        return context


class BootstrapButtonFormMixin(LinkFormMixin, IconFormMixin, ButtonFormMixin):
    require_link = False
    require_icon = False


class BootstrapButtonPlugin(BootstrapButtonMixin, LinkPluginBase):
    module = 'Bootstrap'
    name = _("Button")
    model_mixins = (LinkElementMixin,)
    form = BootstrapButtonFormMixin
    ring_plugin = 'ButtonPlugin'
    DEFAULT_BUTTON_ATTRIBUTES = {'role': 'button'}

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/buttonplugin.js']

    @classmethod
    def get_identifier(cls, instance):
        content = instance.glossary.get('link_content')
        if not content:
            try:
                button_types = dict(ButtonFormMixin.BUTTON_TYPES)
                content = str(button_types[instance.glossary['button_type']])
            except KeyError:
                content = _("Empty")
        return content

    @classmethod
    def get_css_classes(cls, obj):
        css_classes = cls.super(BootstrapButtonPlugin, cls).get_css_classes(obj)
        if obj.glossary.get('stretched_link'):
            css_classes.append('stretched_link')
        return css_classes

    @classmethod
    def get_html_tag_attributes(cls, obj):
        attributes = cls.super(BootstrapButtonPlugin, cls).get_html_tag_attributes(obj)
        attributes.update(cls.DEFAULT_BUTTON_ATTRIBUTES)
        return attributes

    def render(self, context, instance, placeholder):
        context = self.super(BootstrapButtonPlugin, self).render(context, instance, placeholder)
        return context

plugin_pool.register_plugin(BootstrapButtonPlugin)



from collections import OrderedDict
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from cmsplugin_cascade.extra_fields.config import PluginExtraFieldsConfig
from cmsplugin_cascade.bootstrap4.mixins import BootstrapUtilities
from .grid import Breakpoint, Bound


CASCADE_PLUGINS = ['accordion', 'buttons', 'card', 'carousel', 'container', 'embeds', 'icon', 'image', 'jumbotron',
                   'picture', 'tabs']
if 'cms_bootstrap' in settings.INSTALLED_APPS:
    CASCADE_PLUGINS.append('secondary_menu')


def set_defaults(config):
    config.setdefault('bootstrap4', {})
    config['bootstrap4'].setdefault('default_bounds', OrderedDict([
        (Breakpoint.xs, Bound(320, 572)),
        (Breakpoint.sm, Bound(540, 540)),
        (Breakpoint.md, Bound(720, 720)),
        (Breakpoint.lg, Bound(960, 960)),
        (Breakpoint.xl, Bound(1140, 1140)),
    ]))
    config['bootstrap4'].setdefault('fluid_bounds', OrderedDict([
        (Breakpoint.xs, Bound(320, 576)),
        (Breakpoint.sm, Bound(576, 768)),
        (Breakpoint.md, Bound(768, 992)),
        (Breakpoint.lg, Bound(992, 1200)),
        (Breakpoint.xl, Bound(1200, 1980)),
    ]))
    config['bootstrap4'].setdefault('gutter', 30)

    config['plugins_with_extra_mixins'].setdefault('BootstrapAccordionPlugin', BootstrapUtilities(
        BootstrapUtilities.margins,
    ))
    config['plugins_with_extra_mixins'].setdefault('BootstrapAccordionGroupPlugin', BootstrapUtilities(
        BootstrapUtilities.background_and_color,
        BootstrapUtilities.margins,
    ))
    config['plugins_with_extra_mixins'].setdefault('BootstrapCardPlugin', BootstrapUtilities(
        BootstrapUtilities.background_and_color,
        BootstrapUtilities.margins,
    ))
    config['plugins_with_extra_mixins'].setdefault('BootstrapCarouselPlugin', BootstrapUtilities(
        BootstrapUtilities.margins,
    ))
    config['plugins_with_extra_mixins'].setdefault('BootstrapContainerPlugin', BootstrapUtilities(
        BootstrapUtilities.paddings,
    ))
    config['plugins_with_extra_mixins'].setdefault('HeadingPlugin', BootstrapUtilities(
        BootstrapUtilities.margins,
    ))
    config['plugins_with_extra_mixins'].setdefault('HorizontalRulePlugin', BootstrapUtilities(
        BootstrapUtilities.margins,
    ))

    config['plugins_with_extra_fields'].setdefault('BootstrapTabSetPlugin', PluginExtraFieldsConfig(
        css_classes={
            'multiple': True,
            'class_names': ['nav-tabs', 'nav-pills', 'nav-fill', 'nav-justified'],
        },
    ))

    config['plugins_with_extra_render_templates'].setdefault('BootstrapSecondaryMenuPlugin', [
        ('cascade/bootstrap4/secmenu-list-group.html', _("List Group")),
        ('cascade/bootstrap4/secmenu-unstyled-list.html', _("Unstyled List"))
    ])



from django.contrib.admin import site as default_admin_site
from django.contrib import messages

from cms.api import add_plugin
from cms.models.placeholderpluginmodel import PlaceholderReference
from cms.plugin_pool import plugin_pool
from cms.utils import get_language_from_request

from djangocms_text_ckeditor.models import Text
from djangocms_text_ckeditor.utils import plugin_tags_to_id_list, replace_plugin_tags

from cmsplugin_cascade.models import CascadeElement


def serialize_from_placeholder(placeholder, admin_site=default_admin_site):
    """
    Create a serialized representation of all the plugins belonging to the clipboard.
    """
    def populate_data(parent, data):
        for child in plugin_qs.filter(parent=parent).order_by('position'):
            instance, plugin = child.get_plugin_instance(admin_site)
            plugin_type = plugin.__class__.__name__
            try:
                entry = (plugin_type, plugin.get_data_representation(instance), [])
            except AttributeError:
                if isinstance(instance, Text):
                    entry = (plugin_type, {'body': instance.body, 'pk': instance.pk}, [])
                else:
                    continue
            data.append(entry)
            populate_data(child, entry[2])

    data = {'plugins': []}
    plugin_qs = placeholder.cmsplugin_set.all()
    populate_data(None, data['plugins'])
    return data


def deserialize_to_clipboard(request, data):
    """
    Restore clipboard's content by creating plugins from given data.
    """
    def plugins_from_data(placeholder, parent, data):
        for plugin_type, data, children_data in data:
            try:
                plugin_class = plugin_pool.get_plugin(plugin_type)
            except Exception:
                messages.add_message(request, messages.ERROR, "Unable create plugin of type: {}".format(plugin_type))
                continue
            kwargs = dict(data)
            inlines = kwargs.pop('inlines', [])
            shared_glossary = kwargs.pop('shared_glossary', None)
            try:
                instance = add_plugin(placeholder, plugin_class, language, target=parent, **kwargs)
            except Exception:
                messages.add_message(request, messages.ERROR, "Unable to create structure for plugin: {}".format(plugin_class.name))
                continue
            if isinstance(instance, CascadeElement):
                instance.plugin_class.add_inline_elements(instance, inlines)
                instance.plugin_class.add_shared_reference(instance, shared_glossary)

            # for some unknown reasons add_plugin sets instance.numchild to 0,
            # but fixing and save()-ing 'instance' executes some filters in an unwanted manner
            plugins_from_data(placeholder, instance, children_data)

            if isinstance(instance, Text):
                # we must convert the old plugin IDs into the new ones,
                # otherwise links are not displayed
                id_dict = dict(zip(
                    plugin_tags_to_id_list(instance.body),
                    (t[0] for t in instance.get_children().values_list('id'))
                ))
                instance.body = replace_plugin_tags(instance.body, id_dict)
                instance.save()

    language = get_language_from_request(request)

    clipboard = request.toolbar.clipboard
    ref_plugin = clipboard.cmsplugin_set.first()
    if ref_plugin is None:
        # the clipboard is empty
        root_plugin = add_plugin(clipboard, 'PlaceholderPlugin', language, name='clipboard')
    else:
        # remove old entries from the clipboard
        try:
            root_plugin = ref_plugin.cms_placeholderreference
        except PlaceholderReference.DoesNotExist:
            root_plugin = add_plugin(clipboard, 'PlaceholderPlugin', language, name='clipboard')
        else:
            inst, _ = ref_plugin.get_plugin_instance()
            inst.placeholder_ref.get_plugins().delete()
    plugins_from_data(root_plugin.placeholder_ref, None, data['plugins'])



import re
from urllib.parse import urlparse, urlunparse, ParseResult

from django.core.exceptions import ValidationError
from django.forms import widgets
from django.forms.fields import BooleanField, ChoiceField, URLField, Field
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin, EntangledField
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.bootstrap4.plugin_base import BootstrapPluginBase


class YoutubeFormMixin(EntangledModelFormMixin):
    ASPECT_RATIO_CHOICES = [
        ('embed-responsive-21by9', _("Responsive 21:9")),
        ('embed-responsive-16by9', _("Responsive 16:9")),
        ('embed-responsive-4by3', _("Responsive 4:3")),
        ('embed-responsive-1by1', _("Responsive 1:1")),
    ]

    videoid = EntangledField()

    url = URLField(
        label=_("YouTube URL"),
        widget=widgets.URLInput(attrs={'size': 50}),
    )

    aspect_ratio = ChoiceField(
        label=_("Aspect Ratio"),
        choices=ASPECT_RATIO_CHOICES,
        widget=widgets.RadioSelect,
        initial=ASPECT_RATIO_CHOICES[1][0],
    )

    allow_fullscreen = BooleanField(
        label=_("Allow Fullscreen"),
        required=False,
        initial=True,
    )

    autoplay = BooleanField(
        label=_("Autoplay"),
        required=False,
    )

    controls = BooleanField(
        label=_("Display Controls"),
        required=False,
    )

    loop = BooleanField(
        label=_("Enable Looping"),
        required=False,
    )

    rel = BooleanField(
        label=_("Show related"),
        required=False,
    )

    class Meta:
        untangled_fields = ['url']
        entangled_fields = {'glossary': ['videoid', 'aspect_ratio', 'allow_fullscreen', 'autoplay',
                                         'controls', 'loop', 'rel']}

    def __init__(self, *args, **kwargs):
        instance = kwargs.get('instance')
        if instance:
            videoid = instance.glossary.get('videoid')
            if videoid:
                parts = ParseResult('https', 'youtu.be', videoid, '', '', '')
                initial = {'url': urlunparse(parts)}
                kwargs.update(initial=initial)
        super().__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = super().clean()
        url = cleaned_data.get('url')
        if url:
            parts = urlparse(url)
            match = re.search(r'^v=([^&]+)', parts.query)
            if match:
                cleaned_data['videoid'] = match.group(1)
                return cleaned_data
            match = re.search(r'([^/]+)$', parts.path)
            if match:
                cleaned_data['videoid'] = match.group(1)
                return cleaned_data
        raise ValidationError(_("Please enter a valid YouTube URL"))


class BootstrapYoutubePlugin(BootstrapPluginBase):
    """
    Use this plugin to display a YouTube video.
    """
    name = _("You Tube")
    require_parent = False
    parent_classes = ['BootstrapColumnPlugin']
    child_classes = None
    render_template = 'cascade/bootstrap4/youtube.html'
    form = YoutubeFormMixin

    def render(self, context, instance, placeholder):
        context = self.super(BootstrapYoutubePlugin, self).render(context, instance, placeholder)
        query_params = ['autoplay', 'controls', 'loop', 'rel']
        videoid = instance.glossary.get('videoid')
        if videoid:
            query = ['{}=1'.format(key) for key in query_params if instance.glossary.get(key)]
            parts = ParseResult('https', 'www.youtube.com', '/embed/' + videoid, '', '&'.join(query), '')
            context.update({
                'youtube_url': urlunparse(parts),
                'allowfullscreen': 'allowfullscreen' if instance.glossary.get('allow_fullscreen') else '',
            })
        return context

    @classmethod
    def get_css_classes(cls, obj):
        css_classes = cls.super(BootstrapYoutubePlugin, cls).get_css_classes(obj)
        css_classes.append('embed-responsive')
        css_class = obj.glossary.get('aspect_ratio')
        if css_class:
            css_classes.append(css_class)
        return css_classes

    @classmethod
    def get_identifier(cls, obj):
        return obj.glossary.get('videoid', '')

plugin_pool.register_plugin(BootstrapYoutubePlugin)



from django.contrib import admin
from django.contrib.auth import get_user_model
from django.http import HttpResponse, HttpResponseBadRequest
from django.template.response import TemplateResponse
from django.urls import reverse, re_path
from django.utils.translation import gettext_lazy as _, ngettext
from django.utils.html import format_html

from cms.constants import REFRESH_PAGE


class SegmentPluginModelMixin:
    """
    TODO: whenever cmsplugin_cascade drops support for django-CMS < 3.4, this mixin class
    shall be added to the plugin rather than to the model
    """

    def get_context_override(self, request):
        """
        Return a dictionary to override the request context object during evaluation with
        alternative values. Normally this is an empty dict. However, when a staff user overrides
        the segmentation, then update the context with this returned dict.
        """
        return {}

    def render_plugin(self, context=None, placeholder=None, admin=False, processors=None):
        context.update(self.get_context_override(context['request']))
        content = super().render_plugin(context, placeholder, admin, processors)
        context.pop()
        return content


class EmulateUserModelMixin(SegmentPluginModelMixin):
    UserModel = get_user_model()

    def get_context_override(self, request):
        """
        Override the request object with an emulated user.
        """
        context_override = super().get_context_override(request)
        try:
            if request.user.is_staff:
                user = self.UserModel.objects.get(pk=request.session['emulate_user_id'])
                context_override.update(user=user)
        except (self.UserModel.DoesNotExist, KeyError):
            pass
        return context_override


class EmulateUserAdminMixin:
    UserModel = get_user_model()

    @staticmethod
    def populate_toolbar(segmentation_menu, request):
        active = 'emulate_user_id' in request.session
        segmentation_menu.add_sideframe_item(
            _("Emulate User"),
            url=reverse('admin:emulate-users'),
            active=active,
        )
        segmentation_menu.add_ajax_item(
            _("Clear emulations"),
            action=reverse('admin:clear-emulations'),
            on_success=REFRESH_PAGE,
            disabled=not active,
        )

    def get_urls(self):
        return [
            re_path(r'^emulate_users/$', self.admin_site.admin_view(self.emulate_users), name='emulate-users'),
            re_path(r'^emulate_user/(?P<user_id>\d+)/$', self.admin_site.admin_view(self.emulate_user), name='emulate-user'),
            re_path(r'^clear_emulations/$', self.admin_site.admin_view(self.clear_emulations), name='clear-emulations'),
        ] + super().get_urls()

    def emulate_user(self, request, user_id):
        try:
            request.session['emulate_user_id'] = int(user_id)
            return HttpResponse('OK')
        except TypeError as err:
            return HttpResponseBadRequest(err.message)

    def emulate_users(self, request):
        """
        The list view
        """
        def display_as_link(obj):
            try:
                identifier = getattr(user_model_admin, list_display_link)(obj)
            except AttributeError:
                identifier = admin.utils.lookup_field(list_display_link, obj, model_admin=self)[2]
            emulate_user_id = request.session.get('emulate_user_id')
            if emulate_user_id == obj.id:
                return format_html('<strong>{}</strong>', identifier)
            fmtargs = {
                'href': reverse('admin:emulate-user', kwargs={'user_id': obj.id}),
                'identifier': identifier,
            }
            return format_html('<a href="{href}" class="emulate-user">{identifier}</a>', **fmtargs)

        opts = self.UserModel._meta
        app_label = opts.app_label
        user_model_admin = self.admin_site._registry[self.UserModel]
        request._lookup_model = self.UserModel
        list_display_links = user_model_admin.get_list_display_links(request, user_model_admin.list_display)
        # replace first entry in list_display_links by customized method display_as_link
        list_display_link = list_display_links[0]
        try:
            list_display = list(user_model_admin.segmentation_list_display)
        except AttributeError:
            list_display = list(user_model_admin.list_display)
        list_display.remove(list_display_link)
        list_display.insert(0, 'display_as_link')
        display_as_link.allow_tags = True  # TODO: presumably not required anymore since Django-1.9
        try:
            display_as_link.short_description = user_model_admin.identifier.short_description
        except AttributeError:
            display_as_link.short_description = admin.utils.label_for_field(list_display_link, self.UserModel)
        self.display_as_link = display_as_link

        ChangeList = self.get_changelist(request)
        cl = ChangeList(request, self.UserModel, list_display,
            (None,),  # disable list_display_links in ChangeList, instead override that field
            user_model_admin.list_filter,
            user_model_admin.date_hierarchy, user_model_admin.search_fields,
            user_model_admin.list_select_related, user_model_admin.list_per_page,
            user_model_admin.list_max_show_all,
            (),  # disable list_editable
            self,
            None)
        cl.formset = None
        selection_note_all = ngettext('%(total_count)s selected',
            'All %(total_count)s selected', cl.result_count)

        context = {
            'module_name': str(opts.verbose_name_plural),
            'selection_note': _('0 of %(cnt)s selected') % {'cnt': len(cl.result_list)},
            'selection_note_all': selection_note_all % {'total_count': cl.result_count},
            'title': _("Select %(user_model)s to emulate") % {'user_model': opts.verbose_name},
            'is_popup': cl.is_popup,
            'cl': cl,
            'media': self.media,
            'has_add_permission': False,
            'opts': cl.opts,
            'app_label': app_label,
            'actions_on_top': self.actions_on_top,
            'actions_on_bottom': self.actions_on_bottom,
            'actions_selection_counter': self.actions_selection_counter,
            'preserved_filters': self.get_preserved_filters(request),
        }
        return TemplateResponse(request, self.change_list_template or [
            'admin/%s/%s/change_list.html' % (app_label, opts.model_name),
            'admin/%s/change_list.html' % app_label,
            'admin/change_list.html'
        ], context)

    def clear_emulations(self, request):
        request.session.pop('emulate_user_id', None)
        return HttpResponse('OK')



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0024_page_icon_font'),
    ]

    operations = [
        migrations.CreateModel(
            name='TextEditorConfigFields',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50, verbose_name='Name')),
                ('element_type', models.CharField(choices=[('p', 'p'), ('h1', 'h1'), ('h2', 'h2'), ('h3', 'h3'), ('h4', 'h4'), ('h5', 'h5'), ('h6', 'h6'), ('pre', 'pre'), ('address', 'address'), ('div', 'div'), ('span', 'span')], max_length=12, verbose_name='Element Type')),
                ('css_classes', models.CharField(help_text='Freely selectable CSS classnames for this Text-Editor Style, separated by spaces.', max_length=250, verbose_name='CSS classes')),
            ],
            options={
                'verbose_name': 'Text Editor Config',
            },
        ),
    ]



from django.utils.module_loading import import_string
from cmsplugin_cascade import app_settings


LinkPluginBase, LinkFormMixin = (import_string(cls)
    for cls in app_settings.CMSPLUGIN_CASCADE['link_plugin_classes'])



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0013_urlconfrevision'),
        ('cmsplugin_cascade', '0008_sortableinlinecascadeelement'),
    ]

    operations = [
        migrations.CreateModel(
            name='CascadePage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('settings', models.JSONField(blank=True, default={}, help_text='User editable settings for this page.')),
                ('glossary', models.JSONField(blank=True, default={}, help_text='Store for arbitrary page data.')),
                ('extended_object', models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE, to='cms.Page')),
                ('public_extension', models.OneToOneField(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='draft_extension', to='cmsplugin_cascade.CascadePage')),
            ],
            options={
                'db_table': 'cmsplugin_cascade_page',
                'verbose_name': 'Cascade Page Settings',
                'verbose_name_plural': 'Cascade Page Settings',
            },
        ),
    ]



from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Segmentation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
            ],
            options={
                'verbose_name': 'Segmentation',
                'managed': False,
                'verbose_name_plural': 'Segmentation',
            },
            bases=(models.Model,),
        ),
        migrations.AlterField(
            model_name='cascadeelement',
            name='glossary',
            field=models.JSONField(default={}, blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='pluginextrafields',
            name='plugin_type',
            field=models.CharField(max_length=50, verbose_name='Plugin Name', db_index=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='sharablecascadeelement',
            name='glossary',
            field=models.JSONField(default={}, blank=True),
            preserve_default=True,
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0006_bootstrapgallerypluginmodel'),
    ]

    operations = [
    ]



from django.db import migrations, models
import django.db.models.deletion
from cmsplugin_cascade.models import CascadeElement, CascadePage, IconFont


def forwards(apps, schema_editor):
    for cascade_element in CascadeElement.objects.all():
        if cascade_element.plugin_type not in ['FramedIconPlugin', 'TextIconPlugin', 'BootstrapButtonPlugin']:
            continue
        try:
            cms_page = cascade_element.page.get_public_object()
            icon_font = cms_page.cascadepage.icon_font
            if not icon_font:
                continue
        except:
            continue
        if 'icon_font' not in cascade_element.glossary:
            cascade_element.glossary['icon_font'] = icon_font.pk
            cascade_element.save()


def backwards(apps, schema_editor):
    print("Backward migration not implemented")


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0023_iconfont_is_default'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cascadepage',
            name='icon_font',
            field=models.ForeignKey(blank=True, help_text='Deprecated', null=True, on_delete=django.db.models.deletion.SET_NULL, to='cmsplugin_cascade.IconFont', verbose_name='Icon Font'),
        ),
        migrations.RunPython(forwards, reverse_code=backwards),
    ]



import logging
from django.forms import widgets, MultipleChoiceField
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.bootstrap4.grid import Breakpoint
from cmsplugin_cascade.bootstrap4.utils import get_picture_elements, IMAGE_RESIZE_OPTIONS, IMAGE_SHAPE_CHOICES
from cmsplugin_cascade.bootstrap4.fields import BootstrapMultiSizeField
from cmsplugin_cascade.image import ImageFormMixin, ImagePropertyMixin
from cmsplugin_cascade.link.config import LinkPluginBase, LinkFormMixin
from cmsplugin_cascade.link.plugin_base import LinkElementMixin

logger = logging.getLogger('cascade.bootstrap4')


class BootstrapPictureFormMixin(ImageFormMixin):
    responsive_heights = BootstrapMultiSizeField(
        label=_("Adapt Picture Heights"),
        required=False,
        require_all_fields=False,
        allowed_units=['px', '%'],
        initial='100%',
        help_text=_("Heights of picture in percent or pixels for distinct Bootstrap's breakpoints."),
    )

    responsive_zoom = BootstrapMultiSizeField(
        label=_("Adapt Picture Zoom"),
        required=False,
        require_all_fields=False,
        allowed_units=['%'],
        initial=['0%', '0%', '0%', '0%', '0%'],
        help_text=_("Magnification of picture in percent for distinct Bootstrap's breakpoints."),
    )

    resize_options = MultipleChoiceField(
        label=_("Resize Options"),
        choices=IMAGE_RESIZE_OPTIONS,
        widget=widgets.CheckboxSelectMultiple,
        initial=['subject_location', 'high_resolution'],
        help_text = _("Options to use when resizing the image."),
    )

    image_shapes = MultipleChoiceField(
        label=_("Image Shapes"),
        choices=IMAGE_SHAPE_CHOICES,
        widget=widgets.CheckboxSelectMultiple,
        initial=['img-fluid']
    )

    class Meta:
        entangled_fields = {'glossary': ['responsive_heights', 'responsive_zoom', 'resize_options', 'image_shapes']}


class BootstrapPicturePlugin(LinkPluginBase):
    name = _("Picture")
    module = 'Bootstrap'
    parent_classes = ['BootstrapColumnPlugin', 'SimpleWrapperPlugin']
    require_parent = True
    allow_children = False
    raw_id_fields = LinkPluginBase.raw_id_fields + ['image_file']
    model_mixins = (ImagePropertyMixin, LinkElementMixin,)
    admin_preview = False
    ring_plugin = 'PicturePlugin'
    form = type('BootstrapPictureForm', (LinkFormMixin, BootstrapPictureFormMixin), {'require_link': False})
    render_template = 'cascade/bootstrap4/linked-picture.html'
    default_css_class = 'img-fluid'
    default_css_attributes = ['image_shapes']
    html_tag_attributes = {'image_title': 'title', 'alt_tag': 'tag'}
    html_tag_attributes.update(LinkPluginBase.html_tag_attributes)

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/pictureplugin.js']

    def render(self, context, instance, placeholder):
        # image shall be rendered in a responsive context using the picture element
        context = self.super(BootstrapPicturePlugin, self).render(context, instance, placeholder)
        try:
            elements = get_picture_elements(instance)
        except Exception as exc:
            logger.warning("Unable generate picture elements. Reason: {}".format(exc))
        else:
            context.update({
                'instance': instance,
                'is_fluid': True,
                'placeholder': placeholder,
                'elements': elements,
            })
        return context

    @classmethod
    def get_css_classes(cls, obj):
        css_classes = cls.super(BootstrapPicturePlugin, cls).get_css_classes(obj)
        css_class = obj.glossary.get('css_class')
        if css_class:
            css_classes.append(css_class)
        return css_classes

    @classmethod
    def get_identifier(cls, obj):
        try:
            content = str(obj.image)
        except AttributeError:
            content = _("No Picture")
        return mark_safe(content)

    @classmethod
    def sanitize_model(cls, obj):
        sanitized = False
        parent = obj.parent
        if parent:
            while parent.plugin_type != 'BootstrapColumnPlugin':
                parent = parent.parent
            grid_column = parent.get_bound_plugin().get_grid_instance()
            obj.glossary.setdefault('media_queries', {})
            for bp in Breakpoint:
                obj.glossary['media_queries'].setdefault(bp.name, {})
                width = round(grid_column.get_bound(bp).max)
                if obj.glossary['media_queries'][bp.name].get('width') != width:
                    obj.glossary['media_queries'][bp.name]['width'] = width
                    sanitized = True
                if obj.glossary['media_queries'][bp.name].get('media') != bp.media_query:
                    obj.glossary['media_queries'][bp.name]['media'] = bp.media_query
                    sanitized = True
        else:
            logger.warning("PicturePlugin(pk={}) has no ColumnPlugin as ancestor.".format(obj.pk))
            return 
        return sanitized

plugin_pool.register_plugin(BootstrapPicturePlugin)
